#!/usr/bin/env python
# coding=utf-8
"""
@Author : yonas
@Time   : 2020/3/15 下午10:55
@File   : web_API.py
"""
import os, time, json, sys
from run.run_cls import Run_Model_Cls
import tornado.ioloop
import tornado.web
from urllib import parse  # parse.quote 文本->url编码 ' '->'%20'  parse.unquote url编码->文本 '%20'->' '

curr_dir = os.path.dirname(os.path.realpath(__file__))
os.environ['CUDA_VISIBLE_DEVICES'] = '5'

rm_cls = Run_Model_Cls('trans_mhattnpool')
rm_cls.restore('trans_ckpt_10')  # restore to infer


def web_predict(sent, need_cut=True):
    time0 = time.time()
    pred = rm_cls.predict([sent], need_cut=need_cut)[0]
    print('elapsed:', time.time() - time0)
    return pred


class MainHandler(tornado.web.RequestHandler):
    def get(self):
        sent = self.get_argument('sent')
        sent = parse.unquote(sent)  # 你好%20嗯嗯 -> 你好 嗯嗯

        print('sent:', sent)
        ret = web_predict(sent)
        ret = {'result': ret}
        ret_str = json.dumps(ret, ensure_ascii=False)
        print(ret_str)
        self.write(ret_str)
        sys.stdout.flush()

    def post(self):
        body_data = self.request.body
        if isinstance(body_data, bytes):
            body_data = body_data.decode('utf-8')
        args_data = json.loads(body_data)
        sent = args_data.get('sent', None)
        print('sent:', sent)
        if sent is None:
            return
        ret = web_predict(sent)
        ret = {'result': ret}
        ret_str = json.dumps(ret, ensure_ascii=False)
        print(ret_str)
        self.write(ret_str)
        sys.stdout.flush()


def make_app():
    return tornado.web.Application([
        (r"/QZ/predict", MainHandler),
    ])


if __name__ == '__main__':
    DEFAULT_PORT = 8090
    app = make_app()
    app.listen(DEFAULT_PORT)
    tornado.ioloop.IOLoop.current().start()

    # curl localhost:8090/QZ/predict?sent=你好



#!/usr/bin/env python
# coding=utf-8
"""
@Author : yonas
@Time   : 2020/3/15 下午10:54
@File   : example.py
"""
import jieba
import numpy as np
import tensorflow as tf
import os, time, re, sys

curr_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(curr_dir)
sys.path.append(curr_dir + '/..')

import common.utils as utils
from model.cls_model import Model


class Deplpy_CLS_Model():
    def __init__(self, ckpt_name=None, pbmodel_dir=None):
        assert ckpt_name or pbmodel_dir, 'at least have one way'
        self.graph = tf.Graph()
        self.config = tf.ConfigProto(allow_soft_placement=True,
                                     gpu_options=tf.GPUOptions(allow_growth=True),
                                     )
        self.sess = tf.Session(config=self.config, graph=self.graph)

        self.token2id_dct = {
            'word2id': utils.Any2Id.from_file(f'{curr_dir}/../data/word2id.dct', use_line_no=True),
            'label2id': utils.Any2Id.from_file(f'{curr_dir}/../data/label2id.dct', use_line_no=True),
        }
        self.jieba = jieba.Tokenizer()
        self.tokenize = lambda t: self.jieba.lcut(re.sub(r'\s+', '，', t))
        self.cut = lambda t: ' '.join(self.tokenize(t))
        if ckpt_name:
            self.load_from_ckpt_meta(ckpt_name)
        else:
            self.load_from_pbmodel(pbmodel_dir)

        self.id2label = self.token2id_dct['label2id'].get_reverse()

    def load_from_ckpt_meta(self, ckpt_name):
        self.model, self.saver = Model.from_ckpt_meta(ckpt_name, self.sess, self.graph)

    def load_from_pbmodel(self, pbmodel_dir):
        self.model = Model.from_pbmodel(pbmodel_dir, self.sess)

    def predict(self, sent, need_cut=True):
        if need_cut:
            sent = self.cut(sent)
        feed_dict = self.model.create_feed_dict_from_raw([sent], [], self.token2id_dct)
        prob = self.sess.run(self.model.y_prob, feed_dict)[0]
        pred = np.argmax(prob)  # [batch]
        return self.id2label[pred]


if __name__ == '__main__':
    os.environ['CUDA_VISIBLE_DEVICES'] = '5'
    dcm = Deplpy_CLS_Model(ckpt_name=curr_dir + '/../run/cls_ckpt_1/trans_mhattnpool-16-1.50-7.29-0.932-0.156.ckpt-160')
    print(dcm.predict('这个是啥'))

    dcm = Deplpy_CLS_Model(pbmodel_dir=curr_dir + '/../run/cls_pbmodel')
    print(dcm.predict('这个是啥'))



import os, pickle, glob
import tensorflow as tf
import numpy as np

import qiznlp.common.utils as utils
import qiznlp.common.train_helper as train_helper


class Run_Model_Base():
    def __init__(self):
        self.model_name = None
        self.sess = None
        self.graph = None
        self.config = None
        self.saver = None
        self.token2id_dct = None
        self.tokenize = None
        self.cut = None
        self.use_hvd = None
        self.model = None

    def save(self, ckpt_dir, model_name=None, epo=None, global_step=None, info_str=''):
        if hasattr(self, 'hvd_rank') and self.hvd_rank != 0:  # 分布式训练时只需1个进程master来保存
            return
        if not os.path.exists(ckpt_dir): os.makedirs(ckpt_dir)
        if model_name is None: model_name = self.model_name
        if global_step is None: global_step = self.model.global_step
        epo = '' if epo is None else f'{epo}-'

        save_path = f'{ckpt_dir}/{model_name}-{epo}{info_str}.ckpt'

        exist_ckpt_path = glob.glob(f'{ckpt_dir}/{model_name}-{epo}*')
        if exist_ckpt_path:
            ckpt_path = exist_ckpt_path[0].rsplit('.', 1)[0]
            [utils.delete_file(file) for file in [
                ckpt_path + '.index',
                ckpt_path + '.meta',
                ckpt_path + '.data-00000-of-00001'
            ]]

        self.saver.save(self.sess, save_path, global_step=global_step)
        # e.g.
        # mrfn-2-1.23-1.18.ckpt-228.index
        # mrfn-2-1.23-1.18.ckpt-228.meta
        print(f'>>>>>> save ckpt ok! {save_path}')

    def restore(self, ckpt_dir, model_name=None, epo=None, step=None):
        if model_name is None: model_name = self.model_name
        if epo is not None:
            restore_path = glob.glob(f'{ckpt_dir}/{model_name}-{epo}-*.ckpt*')[0].rsplit('.', 1)[0]
        elif step is not None:
            restore_path = glob.glob(f'{ckpt_dir}/{model_name}*.ckpt-{step}.*')[0].rsplit('.', 1)[0]
        else:
            restore_path = tf.train.latest_checkpoint(ckpt_dir)
        self.saver.restore(self.sess, restore_path)
        print(f'<<<<<< restoring ckpt from {restore_path}')

    def delete_ckpt(self, ckpt_dir, model_name=None, epo=None):
        if model_name is None: model_name = self.model_name
        if epo is None:  # 删除所有的epo
            epo = ''
        exist_ckpt_path = glob.glob(f'{ckpt_dir}/{model_name}-{epo}*')
        if exist_ckpt_path:  # 如果存在就删除
            ckpt_path = exist_ckpt_path[0].rsplit('.', 1)[0]
            res = [utils.delete_file(file, verbose=False) for file in [
                ckpt_path + '.index',
                ckpt_path + '.meta',
                ckpt_path + '.data-00000-of-00001']]
            if all(res):
                print(f'delete {ckpt_path} success!')

    def prepare_data(self, data_type, raw_data_file,
                     preprocess_raw_data, batch_size,
                     save_data_prefix,
                     **kwargs):
        index = self.hvd_rank if hasattr(self, 'hvd_rank') else None
        shard = self.hvd_size if hasattr(self, 'hvd_size') else None

        if data_type == 'tfrecord':
            train_tfrecord_file, dev_tfrecord_file, test_tfrecord_file = train_helper.prepare_tfrecord(
                raw_data_file, self.model, self.token2id_dct, self.tokenize,
                preprocess_raw_data_fn=preprocess_raw_data,
                save_data_prefix=save_data_prefix,
                **kwargs,
            )

            # 如果载入不成功将返回None, None
            train_dataset, train_data_size = self.model.load_tfrecord(train_tfrecord_file, batch_size=batch_size, index=index, shard=shard)
            dev_dataset, dev_data_size = self.model.load_tfrecord(dev_tfrecord_file, batch_size=batch_size)
            test_dataset, test_data_size = self.model.load_tfrecord(test_tfrecord_file, batch_size=batch_size)

            with self.graph.as_default():
                train_features = train_dataset.make_one_shot_iterator().get_next() if train_dataset else None
                dev_features = dev_dataset.make_one_shot_iterator().get_next() if dev_dataset else None
                test_features = test_dataset.make_one_shot_iterator().get_next() if test_dataset else None

            def gen_feed_dict(i, epo, mode='train'):
                nonlocal train_features, dev_features
                if mode == 'train':
                    assert train_features
                    features = self.sess.run(train_features)
                    if i == 0 and epo == 1:
                        print('inspect tfrecord features:')
                        for k, v in features.items():
                            print(f'{k}: {v.shape}{v.tolist()[:2]}')
                    return self.model.create_feed_dict_from_features(features, 'train')
                elif mode == 'dev':
                    assert dev_features
                    features = self.sess.run(dev_features)
                    return self.model.create_feed_dict_from_features(features, 'dev')
                elif mode == 'test':
                    assert test_features
                    features = self.sess.run(test_features)
                    return self.model.create_feed_dict_from_features(features, 'test')
                else:
                    raise Exception('unsupport mode type')


        elif data_type == 'pkldata':
            train_pkl_file, dev_pkl_file, test_pkl_file = train_helper.prepare_pkldata(
                raw_data_file, self.model, self.token2id_dct, self.tokenize,
                preprocess_raw_data_fn=preprocess_raw_data,
                save_data_prefix=save_data_prefix,
                **kwargs,
            )
            train_data, dev_data, test_data, train_data_size, dev_data_size, test_data_size = (None,) * 6
            trn_total_ids, dev_total_ids, test_total_ids = (None,) * 3
            if os.path.exists(train_pkl_file):
                train_data = pickle.load(open(train_pkl_file, 'rb'))
                train_data_size = train_data['num_data']
                print(f'loading exist train pkl file ok! {train_pkl_file}')
            if os.path.exists(dev_pkl_file):
                dev_data = pickle.load(open(dev_pkl_file, 'rb'))
                dev_data_size = dev_data['num_data']
                print(f'loading exist dev pkl file ok! {dev_pkl_file}')
                dev_total_ids = list(range(dev_data_size))  # dev 按顺序就行
            if os.path.exists(test_pkl_file):
                test_data = pickle.load(open(test_pkl_file, 'rb'))
                test_data_size = test_data['num_data']
                print(f'loading exist test pkl file ok! {test_pkl_file}')
                test_total_ids = list(range(test_data_size))  # test 按顺序就行

            curr_epo = -1

            def gen_feed_dict(i, epo, mode='train'):
                nonlocal curr_epo, trn_total_ids, dev_total_ids, test_total_ids, train_data, dev_data, test_data
                if mode == 'train':
                    assert train_data
                    if curr_epo != epo:  # 换了一个epo了
                        np.random.seed(epo)
                        trn_total_ids = np.random.permutation(train_data_size)  # 根据不同epo从新打乱数据
                        curr_epo = epo
                    if i == 0 and epo == 1:
                        print('inspect pkl data:')
                        for k in train_data:
                            if k != 'num_data':
                                v = [train_data[k][i] for i in trn_total_ids[:2]]
                                print(f'{k}: {v}')
                    ids = trn_total_ids[i * batch_size:(i + 1) * batch_size]
                    return self.model.create_feed_dict_from_data(train_data, ids, 'train')
                elif mode == 'dev':
                    assert dev_data
                    dev_ids = dev_total_ids[i * batch_size:(i + 1) * batch_size]
                    return self.model.create_feed_dict_from_data(dev_data, dev_ids, 'dev')
                elif mode == 'test':
                    assert test_data
                    test_ids = test_total_ids[i * batch_size:(i + 1) * batch_size]
                    return self.model.create_feed_dict_from_data(test_data, test_ids, 'test')
                else:
                    raise Exception('unsupport mode type')
        else:
            raise Exception('unsupport data type')

        train_epo_steps, dev_epo_steps, test_epo_steps = (None,) * 3
        if train_data_size:
            train_epo_steps = int(np.ceil(train_data_size / batch_size))
        if dev_data_size:
            dev_epo_steps = int(np.ceil(dev_data_size / batch_size))
        if test_data_size:
            test_epo_steps = int(np.ceil(test_data_size / batch_size))
        print('batch_size:', batch_size)
        print('train_data_size:', train_data_size)
        print('train_epo_steps:', train_epo_steps)
        print('dev_data_size:', dev_data_size)
        print('dev_epo_steps:', dev_epo_steps)
        print('test_data_size:', test_data_size)
        print('test_epo_steps:', test_epo_steps)

        return train_epo_steps, dev_epo_steps, test_epo_steps, gen_feed_dict

    def stop_training(self, early_stop_patience, train_info, indicator='dev_acc', greater_is_better=True):
        # e.g. patience=3 第2、3、4epo都低于第1epo, 则停止
        if not train_info:
            return False
        assert indicator in list(train_info.values())[0], f'indicator {indicator} not in train_info'
        patience = early_stop_patience
        epo_list = sorted(train_info.keys())
        if len(epo_list) <= patience:
            return False
        pivot = epo_list[-(patience + 1)]
        flags = []
        for e in epo_list[-patience:]:
            if greater_is_better:  # 指标越大越好,如准确率
                flags.append(train_info[pivot][indicator] >= train_info[e][indicator])  # 第1轮比第2、3、4指标都好则停止
            else:  # 指标越小越好,如loss
                flags.append(train_info[pivot][indicator] <= train_info[e][indicator])  # 第1轮比第2、3、4指标都好则停止
        return all(flags)

    def should_save(self, curr_epo, train_info, indicator='dev_acc', greater_is_better=True):
        if not train_info:
            return True
        if len(train_info) == 1:
            return True
        assert indicator in list(train_info.values())[0], f'indicator {indicator} not in train_info'
        indicator_lst = [train_info[e][indicator] for e in train_info if e != curr_epo]
        best_indicator = max(indicator_lst) if greater_is_better else min(indicator_lst)
        if greater_is_better:
            if train_info[curr_epo][indicator] > best_indicator:
                return True
            else:
                return False
        else:
            if train_info[curr_epo][indicator] < best_indicator:
                return True
            else:
                return False

    def get_best_epo(self, train_info, indicator='dev_acc', greater_is_better=True):
        if not train_info:
            return 0
        if len(train_info) == 1:
            return 1
        assert indicator in list(train_info.values())[0], f'indicator {indicator} not in train_info'
        indicator_lst = [[epo, v[indicator]] for epo, v in train_info.items()]
        indicator_lst.sort(key=lambda e: e[0], reverse=True)  # 优先后面的epo
        indicator_lst.sort(key=lambda e: e[1], reverse=True if greater_is_better else False)
        return indicator_lst[0][0]

    def export_model(self, pbmodel_dir):
        with self.graph.as_default():  # 坑
            builder = tf.saved_model.builder.SavedModelBuilder(pbmodel_dir)

            inputs, outputs = self.model.get_signature_export_model()
            signature_def_inputs = {k: tf.saved_model.utils.build_tensor_info(v) for k, v in inputs.items()}
            signature_def_outputs = {k: tf.saved_model.utils.build_tensor_info(v) for k, v in outputs.items()}

            default_key = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY  # or 'chitchat_predict' 签名
            signature_def_map = {default_key:
                tf.saved_model.signature_def_utils.build_signature_def(
                    inputs=signature_def_inputs,
                    outputs=signature_def_outputs,
                    method_name=tf.saved_model.signature_constants.PREDICT_METHOD_NAME
                )
            }
            builder.add_meta_graph_and_variables(
                self.sess,
                [tf.saved_model.tag_constants.SERVING],
                signature_def_map=signature_def_map
            )
            builder.save(as_text=False)
            print(f'export pb model ok! {pbmodel_dir}')
            return pbmodel_dir



#!/usr/bin/env python
# encoding=utf-8

import os
import shutil

def qiznlp_init():
    cwd = os.getcwd()
    curr_dir = os.path.dirname(__file__)

    print('copying ...')
    shutil.copytree(curr_dir + '/model', cwd + '/model')
    print('copy model-dir finish')
    shutil.copytree(curr_dir + '/run', cwd + '/run')
    os.remove(cwd + '/run/run_base.py')
    print('copy run-dir finish')
    shutil.copytree(curr_dir + '/data', cwd + '/data')
    print('copy data-dir finish')
    # shutil.copytree(curr_dir + '/common', cwd + '/common')
    # print('copy common-dir finish')




import os
import tensorflow as tf
from tensorflow.python.util import nest
import numpy as np

curr_dir = os.path.dirname(os.path.realpath(__file__))

from qiznlp.common.modules.common_layers import shape_list, mask_nonpad_from_embedding, add_timing_signal_1d, get_timing_signal_1d, shift_right, split_heads
from qiznlp.common.modules.embedding import embedding, proj_logits
from qiznlp.common.modules.encoder import transformer_encoder, transformer_decoder, EncDecAttention
from qiznlp.common.modules.birnn import Bi_RNN
from qiznlp.common.modules.beam_search import beam_search, greedy_search, get_state_shape_invariants
import qiznlp.common.utils as utils

conf = utils.dict2obj({
    "vocab_size": 4000,
    "embed_size": 300,
    'hidden_size': 300,
    'num_heads': 6,
    'num_encoder_layers': 6,
    'num_decoder_layers': 6,
    'dropout_rate': 0.2,
    'lr': 1e-3,
    'pretrain_emb': None,
    'beam_size': 5,
    'max_decode_len': 50,
    'eos_id': 2,
})


class Model(object):
    def __init__(self, build_graph=True, **kwargs):
        self.conf = conf
        if build_graph:
            # build placeholder
            self.build_placeholder()
            # build model
            self.model_name = kwargs.get('model_name', 'trans')
            {
                'trans': self.build_model,
                # 'rnn_s2s': self.build_model1,
                'rnn_s2s': self.build_model2,
                # add new here
            }[self.model_name]()
            print(f'model_name: {self.model_name} build graph ok!')

    def build_placeholder(self):
        # placeholder
        # 原则上模型输入输出不变，不需换新model
        self.s1 = tf.placeholder(tf.int32, [None, None], name='s1')
        self.s2 = tf.placeholder(tf.int32, [None, None], name='s2')
        self.dropout_rate = tf.placeholder(tf.float32, name="dropout_rate")

    def build_model(self):
        # embedding  # [batch,len,embed]
        s1_embed, _ = embedding(tf.expand_dims(self.s1, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)
        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)

        # encoder
        encoder_input = s1_embed
        encoder_valid_mask = mask_nonpad_from_embedding(encoder_input)  # [batch,len] 1 for nonpad; 0 for pad
        encoder_input = add_timing_signal_1d(encoder_input)  # add position embedding
        encoder_input = tf.layers.dropout(encoder_input, rate=self.dropout_rate)  # dropout

        encoder_output = transformer_encoder(encoder_input, encoder_valid_mask,
                                             hidden_size=conf.hidden_size,
                                             filter_size=conf.hidden_size * 4,
                                             num_heads=conf.num_heads,
                                             num_encoder_layers=conf.num_encoder_layers,
                                             dropout=self.dropout_rate,
                                             attention_dropout=self.dropout_rate,
                                             relu_dropout=self.dropout_rate,
                                             )

        # decoder
        decoder_input = s2_embed
        decoder_valid_mask = mask_nonpad_from_embedding(decoder_input)  # [batch,len] 1 for nonpad; 0 for pad

        decoder_input = shift_right(decoder_input)  # 用pad当做eos
        decoder_input = add_timing_signal_1d(decoder_input)
        decoder_input = tf.layers.dropout(decoder_input, rate=self.dropout_rate)  # dropout

        decoder_output = transformer_decoder(decoder_input, encoder_output, decoder_valid_mask, encoder_valid_mask,
                                             cache=None,
                                             hidden_size=conf.hidden_size,
                                             filter_size=conf.hidden_size * 4,
                                             num_heads=conf.num_heads,
                                             num_decoder_layers=conf.num_decoder_layers,
                                             dropout=self.dropout_rate,
                                             attention_dropout=self.dropout_rate,
                                             relu_dropout=self.dropout_rate,
                                             )

        logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')

        onehot_s2 = tf.one_hot(self.s2, depth=conf.vocab_size)  # [batch,len,vocab]

        xentropy = tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits, labels=onehot_s2)  # [batch,len]
        weights = tf.to_float(tf.not_equal(self.s2, 0))  # [batch,len] 1 for nonpad; 0 for pad

        loss_num = xentropy * weights  # [batch,len]
        loss_den = weights  # [batch,len]

        loss = tf.reduce_sum(loss_num) / tf.reduce_sum(loss_den)  # scalar
        self.loss = loss
        # self.sent_loss = tf.reduce_sum(loss_num, -1) / tf.reduce_sum(loss_den, -1)  # [batch]

        """ 
        transformer decoder infer 
        """
        batch_size = shape_list(encoder_output)[0]
        # 初始化缓存
        cache = {
            'layer_%d' % layer: {
                # 用以缓存decoder过程前面已计算的k,v
                'k': split_heads(tf.zeros([batch_size, 0, conf.embed_size]), conf.num_heads),
                'v': split_heads(tf.zeros([batch_size, 0, conf.embed_size]), conf.num_heads)
            } for layer in range(conf.num_decoder_layers)
        }
        for layer in range(conf.num_decoder_layers):
            # 对于decoder每层均需与encoder顶层隐状态计算attention,相应均有特定的k,v可缓存
            layer_name = 'layer_%d' % layer
            with tf.variable_scope('decoder/%s/encdec_attention/multihead_attention' % layer_name):
                k_encdec = tf.layers.dense(encoder_output, conf.embed_size, use_bias=False, name='k', reuse=tf.AUTO_REUSE)
                k_encdec = split_heads(k_encdec, conf.num_heads)
                v_encdec = tf.layers.dense(encoder_output, conf.embed_size, use_bias=False, name='v', reuse=tf.AUTO_REUSE)
                v_encdec = split_heads(v_encdec, conf.num_heads)
            cache[layer_name]['k_encdec'] = k_encdec
            cache[layer_name]['v_encdec'] = v_encdec
        cache['encoder_output'] = encoder_output
        cache['encoder_mask'] = encoder_valid_mask

        # position embedding
        position_embedding = get_timing_signal_1d(conf.max_decode_len, conf.embed_size)  # +eos [1,length+1,embed]

        def symbols_to_logits_fn(ids, i, cache):
            ids = ids[:, -1:]  # [batch,1] 截取最后一个
            target_embed, _ = embedding(tf.expand_dims(ids, axis=-1), conf.vocab_size, conf.embed_size, 'share_embedding', reuse=True)  # [batch,1,hidden]

            decoder_input = target_embed + position_embedding[:, i:i + 1, :]  # [batch,1,hidden]

            encoder_output = cache['encoder_output']
            encoder_mask = cache['encoder_mask']

            with tf.variable_scope('', reuse=tf.AUTO_REUSE):
                decoder_output = transformer_decoder(decoder_input, encoder_output, None, encoder_mask,
                                                     cache=cache,  # 注意infer要cache
                                                     hidden_size=conf.embed_size,
                                                     filter_size=conf.embed_size * 4,
                                                     num_heads=6,
                                                     num_decoder_layers=6,
                                                     )
            logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')  # [batch,1,vocab]
            ret = tf.squeeze(logits, axis=1)  # [batch,vocab]
            return ret, cache

        initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

        def greedy_search_wrapper():
            """ Greedy Search """
            decoded_ids, scores = greedy_search(
                symbols_to_logits_fn,
                initial_ids,
                conf.max_decode_len,
                cache=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        def beam_search_wrapper():
            """ Beam Search """
            decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
                symbols_to_logits_fn,
                initial_ids,
                conf.beam_size,
                conf.max_decode_len,
                conf.vocab_size,
                alpha=0,
                states=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        decoded_ids, scores = tf.cond(tf.equal(conf.beam_size, 1), greedy_search_wrapper, beam_search_wrapper)

        self.decoded_ids = tf.identity(decoded_ids, name='decoded_ids')  # [batch,beam/1,len]
        self.scores = tf.identity(scores, name='scores')  # [batch,beam/1]

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    def build_model1(self):
        # embedding
        # [batch,len,embed]
        # pretrained_word_embeddings = np.load(f'{curr_dir}/pretrain_emb_300.npy')
        pretrained_word_embeddings = None
        s1_embed, _ = embedding(tf.expand_dims(self.s1, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=pretrained_word_embeddings)
        s1_mask = mask_nonpad_from_embedding(s1_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s1_seqlen = tf.cast(tf.reduce_sum(s1_mask, axis=-1), tf.int32)  # [batch]

        # encoder
        encoder_input = s1_embed
        encoder_input = tf.layers.dropout(encoder_input, rate=self.dropout_rate)  # dropout

        with tf.variable_scope('birnn_encoder'):
            self.bilstm_encoder1 = Bi_RNN(cell_name='GRUCell', hidden_size=conf.hidden_size, dropout_rate=self.dropout_rate)
            encoder_output, _ = self.bilstm_encoder1(encoder_input, s1_seqlen)  # [batch,len,2hid]

        # decoder

        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=pretrained_word_embeddings)
        s2_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s2_seqlen = tf.cast(tf.reduce_sum(s2_mask, -1), tf.int32)  # [batch]

        decoder_input = s2_embed
        decoder_input = shift_right(decoder_input)  # 用pad当做eos
        decoder_input = tf.layers.dropout(decoder_input, rate=self.dropout_rate)  # dropout

        decoder_rnn = tf.nn.rnn_cell.DropoutWrapper(getattr(tf.nn.rnn_cell, 'GRUCell')(conf.hidden_size),  # GRUCell/LSTMCell
                                                    input_keep_prob=1.0 - self.dropout_rate)

        attention_mechanism = getattr(tf.contrib.seq2seq, 'BahdanauAttention')(
            conf.hidden_size,
            encoder_output,
            memory_sequence_length=s1_seqlen,
            name='BahdanauAttention',
        )
        cell = tf.contrib.seq2seq.AttentionWrapper(decoder_rnn,
                                                   attention_mechanism,
                                                   output_attention=False,
                                                   name='attention_wrapper',
                                                   )

        with tf.variable_scope('decoder'):
            decoder_output, _ = tf.nn.dynamic_rnn(
                cell,
                decoder_input,
                s2_seqlen,
                initial_state=None,  # 默认用0向量初始化
                dtype=tf.float32,
                time_major=False
            )  # 默认scope是rnn e.g.decoder/rnn/kernal

        logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')
        # logits = proj_logits(encoder_output[:,:,:300], conf.embed_size, conf.vocab_size, name='share_embedding')

        onehot_s2 = tf.one_hot(self.s2, depth=conf.vocab_size)  # [batch,len,vocab]

        xentropy = tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits, labels=onehot_s2)  # [batch,len]
        weights = tf.to_float(tf.not_equal(self.s2, 0))  # [batch,len] 1 for nonpad; 0 for pad

        loss_num = xentropy * weights  # [batch,len]
        loss_den = weights  # [batch,len]

        loss = tf.reduce_sum(loss_num) / tf.reduce_sum(loss_den)  # scalar
        self.loss = loss
        # self.sent_loss = tf.reduce_sum(loss_num, -1) / tf.reduce_sum(loss_den, -1)  # [batch]

        """ 
        decoder infer 
        """
        # batch_size = shape_list(encoder_output)[0]
        batch_size = shape_list(encoder_output)[0]
        last_dim = shape_list(encoder_output)[-1]
        tile_encoder_output = tf.tile(tf.expand_dims(encoder_output, 1), [1, conf.beam_size, 1, 1])
        tile_encoder_output = tf.reshape(tile_encoder_output, [batch_size * conf.beam_size, -1, last_dim])
        tile_s1_seqlen = tf.tile(tf.expand_dims(s1_seqlen, 1), [1, conf.beam_size])
        tile_s1_seqlent = tf.reshape(tile_s1_seqlen, [-1])

        # 因为tf.BahdanauAttention在初始时就要指定beam_size来tile memory, 所以这里初始化另外一个专用于推断的beam_size tiled的并共享参数
        # 不过验证还是有问题，不能用reuse=True 报 Variable memory_layer_1/kernel does not exist, or was not created with tf.get_variable()
        # 故不建议使用
        with tf.variable_scope('', reuse=tf.AUTO_REUSE):
            attention_mechanism_decoder = getattr(tf.contrib.seq2seq, 'BahdanauAttention')(
                conf.hidden_size,
                tile_encoder_output,
                memory_sequence_length=tile_s1_seqlent,
                name='BahdanauAttention',
            )
            cell_decoder = tf.contrib.seq2seq.AttentionWrapper(decoder_rnn,
                                                               attention_mechanism_decoder,
                                                               output_attention=False,
                                                               name='attention_wrapper',
                                                               )

        initial_state = cell_decoder.zero_state(batch_size * conf.beam_size, tf.float32)  # 内部会检查batch_size与encoder_output是否一致,需乘beam_size
        
        # 初始化缓存
        # 区分能否设在cache: cache的值在beam_search过程中会expand和merge,需要tensor rank大于1
        cache = {
            'cell_state': initial_state.cell_state,
            'attention': initial_state.attention,
            'alignments': initial_state.alignments,
            'attention_state': initial_state.attention_state,
        }
        unable_cache = {
            'alignment_history': initial_state.alignment_history,
            # 'time': initial_state.time
        }

        # 将cache先变回batch,beam_search过程会expand/merge/gather,使得state是符合batch*beam的
        cache = nest.map_structure(lambda s: s[:batch_size], cache)

        def symbols_to_logits_fn(ids, i, cache):
            nonlocal unable_cache
            ids = ids[:, -1:]
            target = tf.expand_dims(ids, axis=-1)  # [batch,1,1]
            embedding_target, _ = embedding(target, conf.vocab_size, conf.hidden_size, 'share_embedding', reuse=True)
            input = tf.squeeze(embedding_target, axis=1)  # [batch,hid]

            # 合并 cache和unable_cache为state
            state = cell_decoder.zero_state(batch_size * conf.beam_size, tf.float32).clone(
                cell_state=cache['cell_state'],
                attention=cache['attention'],
                alignments=cache['alignments'],
                attention_state=cache['attention_state'],
                alignment_history=unable_cache['alignment_history'],
                # time=unable_cache['time'],
                time=tf.convert_to_tensor(i, dtype=tf.int32),
            )

            with tf.variable_scope('decoder/rnn', reuse=tf.AUTO_REUSE):
                output, state = cell_decoder(input, state)
            # 分开cache和unable_cache
            cache['cell_state'] = state.cell_state
            cache['attention'] = state.attention
            cache['alignments'] = state.alignments
            cache['attention_state'] = state.attention_state
            unable_cache['alignment_history'] = state.alignment_history
            # unable_cache['time'] = state.time
            body_output = output  # [batch,hidden]

            logits = proj_logits(body_output, conf.embed_size, conf.vocab_size, name='share_embedding')
            return logits, cache

        initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

        def greedy_search_wrapper():
            """ Greedy Search """
            decoded_ids, scores = greedy_search(
                symbols_to_logits_fn,
                initial_ids,
                conf.max_decode_len,
                cache=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        def beam_search_wrapper():
            """ Beam Search """
            decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
                symbols_to_logits_fn,
                initial_ids,
                conf.beam_size,
                conf.max_decode_len,
                conf.vocab_size,
                alpha=0,
                states=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        decoded_ids, scores = tf.cond(tf.equal(conf.beam_size, 1), greedy_search_wrapper, beam_search_wrapper)

        self.decoded_ids = tf.identity(decoded_ids, name='decoded_ids')  # [batch,beam/1,len]
        self.scores = tf.identity(scores, name='scores')  # [batch,beam/1]

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    def build_model2(self):
        # embedding
        # [batch,len,embed]
        # pretrained_word_embeddings = np.load(f'{curr_dir}/pretrain_emb_300.npy')
        pretrained_word_embeddings = None
        s1_embed, _ = embedding(tf.expand_dims(self.s1, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=pretrained_word_embeddings)
        s1_mask = mask_nonpad_from_embedding(s1_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s1_seqlen = tf.cast(tf.reduce_sum(s1_mask, axis=-1), tf.int32)  # [batch]

        # encoder
        encoder_input = s1_embed
        encoder_input = tf.layers.dropout(encoder_input, rate=self.dropout_rate)  # dropout

        with tf.variable_scope('birnn_encoder'):
            self.bilstm_encoder1 = Bi_RNN(cell_name='GRUCell', hidden_size=conf.hidden_size, dropout_rate=self.dropout_rate)
            encoder_output, _ = self.bilstm_encoder1(encoder_input, s1_seqlen)  # [batch,len,2hid]

        batch_size = shape_list(encoder_input)[0]
        
        # decoder
        decoder_rnn = getattr(tf.nn.rnn_cell, 'GRUCell')(conf.hidden_size)  # GRUCell/LSTMCell
        encdec_atten = EncDecAttention(encoder_output, s1_seqlen, conf.hidden_size)

        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=pretrained_word_embeddings)
        s2_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s2_seqlen = tf.cast(tf.reduce_sum(s2_mask, -1), tf.int32)  # [batch]

        decoder_input = s2_embed
        decoder_input = shift_right(decoder_input)  # 用pad当做eos
        decoder_input = tf.layers.dropout(decoder_input, rate=self.dropout_rate)  # dropout

        init_decoder_state = decoder_rnn.zero_state(batch_size, tf.float32)
        # init_decoder_state = tf.nn.rnn_cell.LSTMStateTuple(c, h)

        time_step = tf.constant(0, dtype=tf.int32)
        rnn_output = tf.zeros([batch_size, 0, conf.hidden_size])
        context_output = tf.zeros([batch_size, 0, conf.hidden_size * 2])  # 注意力

        def loop_condition(time_step, *_):
            return tf.less(time_step, tf.reduce_max(s2_seqlen))

        def loop_body(time_step, prev_rnn_state, rnn_output, context_output):
            # attention
            s = prev_rnn_state if isinstance(decoder_rnn, tf.nn.rnn_cell.GRUCell) else prev_rnn_state.h
            context = encdec_atten(s)  # [batch,hidden]
            context_output = tf.concat([context_output, tf.expand_dims(context, axis=1)], axis=1)

            # construct rnn input
            rnn_input = tf.concat([decoder_input[:, time_step, :], context], axis=-1)  # [batch,hidden+]  use attention
            # rnn_input = decoder_input[:, time_step, :]  # [batch,hidden]  not use attention

            # run rnn
            current_output, rnn_state = decoder_rnn(rnn_input, prev_rnn_state)

            # append to output bucket via length dim
            rnn_output = tf.concat([rnn_output, tf.expand_dims(current_output, axis=1)], axis=1)

            return time_step + 1, rnn_state, rnn_output, context_output

        # start loop
        final_time_step, final_state, rnn_output, context_output = tf.while_loop(
            loop_condition,
            loop_body,
            loop_vars=[time_step, init_decoder_state, rnn_output, context_output],
            shape_invariants=[
                tf.TensorShape([]),
                nest.map_structure(get_state_shape_invariants, init_decoder_state),
                tf.TensorShape([None, None, conf.hidden_size]),
                tf.TensorShape([None, None, conf.hidden_size * 2]),
            ])
        # body_output = tf.concat([rnn_output, context_output], axis=-1)
        # body_output = tf.layers.dense(body_output, self.hidden_size, activation=tf.nn.tanh, use_bias=True, name='body_output_layer')
        decoder_output = rnn_output

        logits = proj_logits(decoder_output, conf.embed_size, conf.vocab_size, name='share_embedding')
        # logits = proj_logits(encoder_output[:,:,:300], conf.embed_size, conf.vocab_size, name='share_embedding')

        onehot_s2 = tf.one_hot(self.s2, depth=conf.vocab_size)  # [batch,len,vocab]

        xentropy = tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits, labels=onehot_s2)  # [batch,len]
        weights = tf.to_float(tf.not_equal(self.s2, 0))  # [batch,len] 1 for nonpad; 0 for pad

        loss_num = xentropy * weights  # [batch,len]
        loss_den = weights  # [batch,len]

        loss = tf.reduce_sum(loss_num) / tf.reduce_sum(loss_den)  # scalar
        self.loss = loss
        # self.sent_loss = tf.reduce_sum(loss_num, -1) / tf.reduce_sum(loss_den, -1)  # [batch]


        # decoder infer 
        cache = {'state': decoder_rnn.zero_state(batch_size, tf.float32)}
        def symbols_to_logits_fn(ids, i, cache):
            # ids [batch,length]
            pred_target = ids[:, -1:]  # 截取最后一个  [batch,1]
            embed_target, _ = embedding(tf.expand_dims(pred_target, axis=-1), conf.vocab_size, conf.embed_size, 'share_embedding')  # [batch,length,embed]
            decoder_input = tf.squeeze(embed_target, axis=1)  # [batch,embed]

            # if use attention
            s = cache['state'] if isinstance(decoder_rnn, tf.nn.rnn_cell.GRUCell) else cache['state'].h
            context = encdec_atten(s, beam_size=conf.beam_size)  # [batch,hidden]
            decoder_input = tf.concat([decoder_input, context], axis=-1)  # [batch,hidden+]

            # run rnn
            # with tf.variable_scope('rnn', reuse=tf.AUTO_REUSE):
            decoder_output, cache['state'] = decoder_rnn(decoder_input, cache['state'])

            logits = proj_logits(decoder_output, conf.hidden_size, conf.vocab_size, name='share_embedding')

            return logits, cache

        initial_ids = tf.zeros([batch_size], dtype=tf.int32)  # <pad>为<sos>

        def greedy_search_wrapper():
            """ Greedy Search """
            decoded_ids, scores = greedy_search(
                symbols_to_logits_fn,
                initial_ids,
                conf.max_decode_len,
                cache=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        def beam_search_wrapper():
            """ Beam Search """
            decoded_ids, scores = beam_search(  # [batch,beam,len] [batch,beam]
                symbols_to_logits_fn,
                initial_ids,
                conf.beam_size,
                conf.max_decode_len,
                conf.vocab_size,
                alpha=0,
                states=cache,
                eos_id=conf.eos_id,
            )
            return decoded_ids, scores

        decoded_ids, scores = tf.cond(tf.equal(conf.beam_size, 1), greedy_search_wrapper, beam_search_wrapper)

        self.decoded_ids = tf.identity(decoded_ids, name='decoded_ids')  # [batch,beam/1,len]
        self.scores = tf.identity(scores, name='scores')  # [batch,beam/1]

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    @classmethod
    def sent2ids(cls, sent, word2id, max_word_len=None):
        # sent 已分好词 ' '隔开
        # 形成batch时才动态补齐长度
        words = sent.split(' ')
        token_ids = [word2id.get(word, word2id['<unk>']) for word in words]
        if max_word_len:
            token_ids = token_ids[:max_word_len - 1]
        token_ids.append(word2id['<eos>'])
        # token_ids = pad_sequences([token_ids], padding='post', maxlen=max_word_len)[0]
        return token_ids  # [len]

    def create_feed_dict_from_data(self, data, ids, mode='train'):
        # data:数据已经转为id, data不同字段保存该段字段全量数据
        batch_s1 = [data['s1'][i] for i in ids]
        batch_s2 = [data['s2'][i] for i in ids]
        if len(set([len(e) for e in batch_s1])) != 1:  # 长度不等
            batch_s1 = utils.pad_sequences(batch_s1, padding='post')
        if len(set([len(e) for e in batch_s2])) != 1:  # 长度不等
            batch_s2 = utils.pad_sequences(batch_s2, padding='post')
        feed_dict = {
            self.s1: batch_s1,
            self.s2: batch_s2,
        }
        if mode == 'train': feed_dict['num'] = len(batch_s1)
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_features(self, features, mode='train'):
        # feature:tfrecord数据的example, 每个features的不同字段包括该字段一个batch数据
        feed_dict = {
            self.s1: features['s1'],
            self.s2: features['s2'],
        }
        if mode == 'train': feed_dict['num'] = len(features['s1'])
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_raw(self, batch_s1, batch_s2, token2id_dct, mode='infer'):
        word2id = token2id_dct['word2id']

        feed_s1 = [self.sent2ids(s1, word2id) for s1 in batch_s1]
        if len(set([len(e) for e in feed_s1])) != 1:  # 长度不等
            feed_s1 = utils.pad_sequences(feed_s1, padding='post')

        feed_dict = {
            self.s1: np.array(feed_s1),
        }
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        if mode == 'infer':
            return feed_dict

        if mode in ['train', 'dev']:
            feed_s2 = []
            for s2 in batch_s2:
                s2_np = self.sent2ids(s2, word2id)
                feed_s2.append(s2_np)
            if len(set([len(e) for e in feed_s2])) != 1:  # 长度不等
                feed_s2 = utils.pad_sequences(feed_s2, padding='post')

            feed_dict[self.s2] = np.array(feed_s2)
            return feed_dict

        raise ValueError(f'mode type {mode} not support')

    @classmethod
    def generate_data(cls, file, token2id_dct):
        word2id = token2id_dct['word2id']
        data = {
            's1': [],
            's2': [],
        }
        with open(file, 'r', encoding='U8') as f:
            for i, line in enumerate(f):
                item = line.strip().split('\t')
                if len(item) != 2:
                    print('error item:', repr(line))
                    continue
                s1 = item[0]
                s2 = item[1]
                s1_ids = cls.sent2ids(s1, word2id, max_word_len=50)
                s2_ids = cls.sent2ids(s2, word2id, max_word_len=50)
                if i < 5:  # check
                    print(f'check {i}:')
                    print(f'{s1} -> {s1_ids}')
                    print(f'{s2} -> {s2_ids}')
                data['s1'].append(s1_ids)
                data['s2'].append(s2_ids)
        data['num_data'] = len(data['s1'])
        return data

    @classmethod
    def generate_tfrecord(cls, file, token2id_dct, tfrecord_file):
        from qiznlp.common.tfrecord_utils import items2tfrecord
        word2id = token2id_dct['word2id']

        def items_gen():
            with open(file, 'r', encoding='U8') as f:
                for i, line in enumerate(f):
                    item = line.strip().split('\t')
                    if len(item) != 2:
                        print('error item:', repr(line))
                        continue
                    try:
                        s1 = item[0]
                        s2 = item[1]
                        s1_ids = cls.sent2ids(s1, word2id, max_word_len=50)
                        s2_ids = cls.sent2ids(s2, word2id, max_word_len=50)
                        if i < 5:  # check
                            print(f'check {i}:')
                            print(f'{s1} -> {s1_ids}')
                            print(f'{s2} -> {s2_ids}')
                        d = {
                            's1': s1_ids,
                            's2': s2_ids,
                        }
                        yield d
                    except:
                        continue

        count = items2tfrecord(items_gen(), tfrecord_file)
        return count

    @classmethod
    def load_tfrecord(cls, tfrecord_file, batch_size=128, index=None, shard=None):
        from qiznlp.common.tfrecord_utils import tfrecord2dataset, exist_tfrecord_file
        if not exist_tfrecord_file(tfrecord_file):
            return None, None
        feat_dct = {
            # 's1': tf.FixedLenFeature([50], tf.int64),
            's1': tf.VarLenFeature(tf.int64),
            's2': tf.VarLenFeature(tf.int64),
        }
        dataset, count = tfrecord2dataset(tfrecord_file, feat_dct, batch_size=batch_size, auto_pad=True, index=index, shard=shard)
        return dataset, count

    def get_signature_export_model(self):
        inputs_dct = {
            's1': self.s1,
            'dropout_rate': self.dropout_rate,
        }
        outputs_dct = {
            'decoded_ids': self.decoded_ids,
            'scores': self.scores,
        }
        return inputs_dct, outputs_dct

    @classmethod
    def get_signature_load_pbmodel(cls):
        inputs_lst = ['s1', 'dropout_rate']
        outputs_lst = ['decoded_ids', 'scores']
        return inputs_lst, outputs_lst

    @classmethod
    def from_pbmodel(cls, pbmodel_dir, sess):
        meta_graph_def = tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], pbmodel_dir)  # 从pb模型载入graph和variable,绑定到sess
        signature = meta_graph_def.signature_def  # 取出signature_def
        default_key = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY  # or 'chitchat_predict' 签名
        inputs_lst, output_lst = cls.get_signature_load_pbmodel()  # 类方法
        pb_dict = {}
        for k in inputs_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].inputs[k].name)  # 从signature中获取输入输出的tensor_name,并从graph中取出
        for k in output_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].outputs[k].name)
        model = cls(build_graph=False)  # 里面不再构造图
        for k, v in pb_dict.items():
            setattr(model, k, v)  # 绑定必要的输入输出到实例
        return model

    @classmethod
    def from_ckpt_meta(cls, ckpt_name, sess, graph):
        with graph.as_default():
            saver = tf.train.import_meta_graph(ckpt_name + '.meta', clear_devices=True)
            sess.run(tf.global_variables_initializer())

        model = cls(build_graph=False)  # 里面不再构造图
        # 绑定必要的输入输出到实例
        model.s1 = graph.get_tensor_by_name('s1:0')
        # model.s2 = graph.get_tensor_by_name('s2:0')
        model.dropout_rate = graph.get_tensor_by_name('dropout_rate:0')
        model.decoded_ids = graph.get_tensor_by_name('decoded_ids:0')
        model.scores = graph.get_tensor_by_name('scores:0')

        saver.restore(sess, ckpt_name)
        print(f':: restore success! {ckpt_name}')
        return model, saver



#!/usr/bin/env python
# coding=utf-8
import numpy as np
import tensorflow as tf


def mask_nonzero(labels):
    """ mask: Assign weight 1.0(true)
        nonmask: Assign weight 0.0(false)
        mask if value is not 0.0
    """
    return tf.to_float(tf.not_equal(labels, 0))


def mask_nonpad_from_ids(ids):
    """ ids: [batch,length]
        Assign 1.0(true) for non pad, 0.0(false) for pad(id=0, all emb element is 0)
        return [batch,length]
    """
    return tf.to_float(tf.not_equal(ids, 0))


def mask_nonpad_from_embedding(emb):
    """ emb: [batch,length,embed]
        Assign 1.0(true) for non pad, 0.0(false) for pad(id=0, all emb element is 0)
        return [batch,length]
    """
    return mask_nonzero(tf.reduce_sum(tf.abs(emb), axis=-1))


def length_from_embedding(emb):
    """ emb: [batch,length,embed]
        return [batch]
    """
    length = tf.reduce_sum(mask_nonpad_from_embedding(emb), axis=-1)
    length = tf.cast(length, tf.int32)
    return length


def length_from_ids(ids):
    """ ids: [batch,length,1]
        return [batch]
    """
    weight_ids = mask_nonzero(ids)
    length = tf.reduce_sum(weight_ids, axis=[1, 2])
    length = tf.cast(length, tf.int32)
    return length


def shift_right(x, pad_value=None):
    """ shift the second dim of x right by one
        decoder中对target进行右偏一位作为decode_input
        x [batch,length,embed]
        pad_value [tile_batch,1,pad_embed] or
        pad_value [pad_embed]
    """
    if pad_value is not None:
        if len(shape_list(x)) == 1:
            pad_embed = tf.reshape([1, 1, -1])
            pad_embed = tf.tile(pad_embed, [shape_list(x)[0], 1, 1])
            shifted_x = tf.concat([pad_embed, x], axis=1)[:, :-1, :]  # length维度左边补pad_embed
        else:
            shifted_x = tf.concat([pad_value, x], axis=1)[:, :-1, :]  # length维度左边补pad_embed
    else:
        shifted_x = tf.pad(x, [[0, 0], [1, 0], [0, 0]])[:, :-1, :]  # length维度左边补pad_embed [0,0...]
    return shifted_x


def shape_list(x):
    """ return list of dims, statically where possible """
    x = tf.convert_to_tensor(x)
    # if unknown rank, return dynamic shape 如果秩都不知道
    if x.get_shape().dims is None:
        return tf.shape(x)
    static = x.get_shape().as_list()
    shape = tf.shape(x)
    ret = []
    for i in range(len(static)):
        dim = shape[i] if static[i] is None else static[i]
        ret.append(dim)
    return ret


def cast_like(x, y):
    """ cast x to y's dtype, if necessary """
    x = tf.convert_to_tensor(x)
    y = tf.convert_to_tensor(y)
    if x.dtype.base_dtype == y.dtype.base_dtype:
        return x
    else:
        return tf.cast(x, y.dtype)


def log_prob_from_logits(logits, axis=-1):
    return logits - tf.reduce_logsumexp(logits, axis=axis, keepdims=True)


def sample_with_temperature(logits, temperature):
    """ 0.0:argmax 1.0:sampling >1.0:random """
    # logits [batch,length,vocab]
    # ret [batch,length]
    if temperature == 0.0:
        return tf.argmax(logits, axis=-1)
    else:
        assert temperature > 0.0
        logits_shape = shape_list(logits)
        reshape_logits = tf.reshape(logits, [-1, logits_shape[-1]]) / temperature
        choices = tf.multinomial(reshape_logits, 1)  # 仅采样1个 该方式只支持2-D
        choices = tf.reshape(choices, logits_shape[:-1])
        return choices


def batch_normalization(x, training, name):
    # with tf.variable_scope(name, reuse=)
    bn_train = tf.layers.batch_normalization(x, training=True, reuse=None, name=name)
    bn_inference = tf.layers.batch_normalization(x, training=False, reuse=True, name=name)
    z = tf.cond(training, lambda: bn_train, lambda: bn_inference)
    return z


def dropout_with_broadcast_dims(x, keep_prob, broadcast_dims=None, **kwargs):
    """封装dropout函数,broadcast_dims对应dropout的noise_shape"""
    assert 'noise_shape' not in kwargs
    if broadcast_dims:
        x_shape = tf.shape(x)
        ndims = len(x.get_shape())
        broadcast_dims = [dim + ndims if dim < 0 else dim for dim in broadcast_dims]  # allow dim like -1 as well
        kwargs['noise_shape'] = [1 if i in broadcast_dims else x_shape[i] for i in range(ndims)]  # 类似[1,length,hidden]
    return tf.nn.dropout(x, keep_prob, **kwargs)


def dropout_no_scaling(x, keep_prob):
    """ 不进行放缩的drop, 用以在token上 """
    if keep_prob == 1.0:
        return x
    mask = tf.less(tf.random_uniform(tf.shape(x)), keep_prob)
    mask = cast_like(mask, x)
    return x * mask


def layer_norm(x, epsilon=1e-06):
    """ layer norm """
    filters = shape_list(x)[-1]
    with tf.variable_scope('layer_norm', values=[x], reuse=None):
        scale = tf.get_variable('layer_norm_scale', [filters], initializer=tf.ones_initializer())
        bias = tf.get_variable('layer_norm_bias', [filters], initializer=tf.zeros_initializer())
    epsilon, scale, bias = [cast_like(t, x) for t in [epsilon, scale, bias]]  # 写法独特

    mean = tf.reduce_mean(x, axis=-1, keepdims=True)  # mu
    variance = tf.reduce_mean(tf.square(x - mean), axis=-1, keepdims=True)  # sigma
    norm_x = (x - mean) * tf.rsqrt(variance + epsilon)  # (x-mu)*sigma
    ret = norm_x * scale + bias
    return ret


def layer_prepostprocess(previous_value, x, sequence, dropout_rate, dropout_broadcast_dims=None):
    """ apply a sequence of function to the input or output of a layer
        a: add previous_value
        n: apply normalization
        d: apply dropout
        for example, if sequence=='dna', then the output is: previous_value + normalize(dropout(x))
    """
    with tf.variable_scope('layer_prepostprocess'):
        if sequence == 'none':
            return x
        for c in sequence:
            if c == 'a':
                x += previous_value  # residual
            elif c == 'n':
                x = layer_norm(x)  # LN
            else:
                assert c == 'd', 'unknown sequence step %s' % c
                x = dropout_with_broadcast_dims(x, 1.0 - dropout_rate, broadcast_dims=dropout_broadcast_dims)  # dropout
        return x


def layer_preprocess(x, dropout_rate=0., layer_preprocess_sequence='n'):
    assert 'a' not in layer_preprocess_sequence, 'no residual connections allowed in preprocess sequence'
    return layer_prepostprocess(None, x, sequence=layer_preprocess_sequence,
                                dropout_rate=dropout_rate)


def layer_postprecess(previous_value, x, dropout_rate=0., layer_postprocess_sequence='da'):
    return layer_prepostprocess(previous_value, x, sequence=layer_postprocess_sequence,
                                dropout_rate=dropout_rate)


def split_heads(x, num_heads):
    """ x [batch,length,hidden]
        ret [batch,num_heads,length,hidden/num_heads]
    """
    x_shape = shape_list(x)
    last_dim = x_shape[-1]
    if isinstance(last_dim, int) and isinstance(num_heads, int):
        assert last_dim % num_heads == 0
    x = tf.reshape(x, x_shape[:-1] + [num_heads, last_dim // num_heads])
    x = tf.transpose(x, [0, 2, 1, 3])
    return x


def combine_heads(x):
    """ Inverse of split_heads
        x [batch,num_heads,length,hidden/num_head]
        ret [batch,length,hidden]
    """
    x = tf.transpose(x, [0, 2, 1, 3])
    x_shape = shape_list(x)
    a, b = x_shape[-2:]
    x = tf.reshape(x, x_shape[:-2] + [a * b])
    return x


def compute_qkv(query_antecedent, memory_antecedent, total_key_depth, total_value_depth):
    """total_depth 包括了所有head的depth"""
    if memory_antecedent is None:
        memory_antecedent = query_antecedent
    q = tf.layers.dense(query_antecedent, total_key_depth, use_bias=False, name='q')
    k = tf.layers.dense(memory_antecedent, total_key_depth, use_bias=False, name='k')
    v = tf.layers.dense(memory_antecedent, total_value_depth, use_bias=False, name='v')
    return q, k, v


def dot_product_attention(q, k, v, bias, dropout_rate=0.0, dropout_broadcast_dims=None, name='dot_product_attention'):
    """ """
    with tf.variable_scope(name, values=[q, k, v]):
        logits = tf.matmul(q, k, transpose_b=True)  # [batch,num_heads,length_q,length_kv]
        if bias is not None:
            bias = cast_like(bias, logits)
        logits += bias
        weights = tf.nn.softmax(logits, name='attention_weight')  # [batch,num_heads,length_q,length_kv]
        if dropout_rate != 0:
            # drop out attention links for each head
            weights = dropout_with_broadcast_dims(weights, 1.0 - dropout_rate, broadcast_dims=dropout_broadcast_dims)
        # v [batch,num_heads,length_kv,hidden/num_heads]
        ret = tf.matmul(weights, v)  # [batch,num_heads,length_q,hidden/num_heads]
        return ret


def multihead_attention(query_antecedent, memory_antecedent, bias, total_key_depth, total_value_depth, output_depth,
                        num_heads, dropout_rate, cache=None, name='multihead_attention', dropout_broadcast_dims=None):
    """ """
    assert total_key_depth % num_heads == 0
    assert total_value_depth % num_heads == 0
    with tf.variable_scope(name, values=[query_antecedent, memory_antecedent]):
        if cache is None or memory_antecedent is None:  # training or self_attention in inferring
            q, k, v = compute_qkv(query_antecedent, memory_antecedent, total_key_depth, total_value_depth)
        if cache is not None:  # inferring时有cache, 此时query_antecedent均为[batch,1,hidden]
            assert bias is not None, 'Bias required for caching'

            if memory_antecedent is not None:  # encode-decode attention 使用cache
                q = tf.layers.dense(query_antecedent, total_key_depth, use_bias=False, name='q')
                k = cache['k_encdec']
                v = cache['v_encdec']
            else:  # decode self_attention 得到k,v需存到cache
                k = split_heads(k, num_heads)  # [batch,num_heads,length,hidden/num_heads]
                v = split_heads(v, num_heads)
                k = cache['k'] = tf.concat([cache['k'], k], axis=2)
                v = cache['v'] = tf.concat([cache['v'], v], axis=2)
        q = split_heads(q, num_heads)
        if cache is None:
            k = split_heads(k, num_heads)
            v = split_heads(v, num_heads)
        key_depth_per_head = total_key_depth // num_heads
        q = q * key_depth_per_head ** -0.5  # scale

        x = dot_product_attention(q, k, v, bias, dropout_rate, dropout_broadcast_dims=dropout_broadcast_dims)

        x = combine_heads(x)

        x.set_shape(x.shape.as_list()[:-1] + [total_key_depth])  # set last dim specifically

        x = tf.layers.dense(x, output_depth, use_bias=False, name='output_transform')
        return x


def attention_bias_lower_triangle(length):
    """ 下三角矩阵 """
    band = tf.matrix_band_part(tf.ones([length, length]), -1, 0)  # [length,length] 下三角矩阵,下三角均为1,上三角均为0
    # [[1,0,0],
    #  [1,1,0],
    #  [1,1,1]] float
    band = tf.reshape(band, [1, 1, length, length])
    band = -1e9 * (1.0 - band)
    # [[0,-1e9,-1e9],
    #  [0,0,-1e9],
    #  [0,0,0]]
    return band


def add_timing_signal_1d(x, min_timescale=1.0, max_timescale=1.0e4, start_index=0):
    """ use for position embedding """
    length = shape_list(x)[1]
    hidden_size = shape_list(x)[2]
    signal = get_timing_signal_1d(length, hidden_size, min_timescale, max_timescale, start_index)
    return x + signal


def get_timing_signal_1d(length, hidden_size, min_timescale=1.0, max_timescale=1.0e4, start_index=0):
    """ use for calculate position embedding """
    position = tf.to_float(tf.range(length) + start_index)
    num_timescales = hidden_size // 2
    log_timescales_increment = np.log(float(max_timescale) / float(min_timescale)) / (tf.to_float(num_timescales - 1))
    inv_timescales = min_timescale * tf.exp(tf.to_float(tf.range(num_timescales)) * -log_timescales_increment)
    scaled_time = tf.expand_dims(position, 1) * tf.expand_dims(inv_timescales, 0)
    signal = tf.concat([tf.sin(scaled_time), tf.cos(scaled_time)], axis=1)
    signal = tf.pad(signal, [[0, 0], [0, tf.mod(hidden_size, 2)]])
    signal = tf.reshape(signal, [1, length, hidden_size])
    return signal  # [1,len,hidden]


def transformer_ffn_layer(x, filter_size, hidden_size, relu_dropout=0., pad_mask=None):
    original_shape = shape_list(x)  # [batch,length,hidden]

    if pad_mask is not None:
        """ remove pad """
        flat_pad_mask = tf.reshape(pad_mask, [-1])  # [batch*length]
        flat_nonpad_ids = tf.to_int32(tf.where(tf.equal(flat_pad_mask, 0)))

        # flat x
        x = tf.reshape(x, tf.concat([[-1], original_shape[2:]], axis=0))  # [batch*length,hidden]
        # remove pad
        x = tf.gather_nd(x, flat_nonpad_ids)  # [batch*length-,hidden]

    h = tf.layers.dense(x, filter_size, use_bias=True, activation=tf.nn.relu, name='conv1')

    if relu_dropout != 0.:
        h = dropout_with_broadcast_dims(h, 1.0 - relu_dropout, broadcast_dims=None)

    o = tf.layers.dense(h, hidden_size, activation=None, use_bias=True, name='conv2')

    if pad_mask is not None:
        """ restore pad """
        o = tf.scatter_nd(  # 将updates中对应的值填充到indices指定的索引中，空的位置会用0代替，刚好代表pad
            indices=flat_nonpad_ids,
            updates=o,
            shape=tf.concat([tf.shape(flat_pad_mask)[:1], tf.shape(o)[1:]], axis=0)
        )

        o = tf.reshape(o, original_shape)
    return o


def sigmoid_cross_entropy(logits, labels):
    loss = tf.nn.sigmoid_cross_entropy_with_logits(logits=logits,
                                                   labels=cast_like(labels, logits))
    loss = tf.reduce_mean(loss)
    return loss


def softmax_cross_entropy(logits, labels):
    loss = tf.nn.softmax_cross_entropy_with_logits_v2(logits=logits,
                                                      labels=cast_like(labels, logits))
    loss = tf.reduce_mean(loss)
    return loss


def am_softmax_loss(logits, labels, margin=0.35, scale=30.):
    labels = tf.cast(labels, tf.float32)
    logits = labels * (logits - margin) + (1 - labels) * logits
    logits *= scale
    loss = softmax_cross_entropy(logits, labels)
    return loss


def margin_loss(logits, labels):
    # logits = tf.nn.softmax(logits)
    labels = tf.cast(labels, tf.float32)
    loss = labels * tf.square(tf.maximum(0., 0.9 - logits)) + \
           0.25 * (1.0 - labels) * tf.square(tf.maximum(0., logits - 0.1))
    loss = tf.reduce_mean(tf.reduce_sum(loss, axis=1))
    return loss


def l1_loss(logits, labels):
    return tf.reduce_mean(tf.abs(logits - labels))


def l2_loss(logits, labels):
    # tf.nn.l2_loss()?
    return tf.reduce_mean(tf.square(logits - labels))


def hinge_loss(neg_logits, pos_logits, margin, is_distance):
    if is_distance:
        loss = tf.reduce_mean(tf.maximum(margin + pos_logits - neg_logits, 0.0))
    else:
        loss = tf.reduce_mean(tf.maximum(margin - pos_logits + neg_logits, 0.0))
    return loss


def improved_triplet_loss(neg_logits, pos_logits, margin, margin_pos, is_distance):
    if is_distance:
        loss = tf.reduce_mean(tf.maximum(margin + pos_logits - neg_logits, 0.0) +
                              tf.square(1 - neg_logits))
        # tf.square(pos_logits))
    else:
        loss = tf.reduce_mean(tf.maximum(margin - pos_logits + neg_logits, 0.0) +
                              tf.square(neg_logits))
        # tf.square(1 - pos_logits))
    return loss


def focal_loss_sigmoid(labels, logits, alpha=0.25, gamma=2):
    """
    Computer focal loss for binary classification
    Args:
      labels: A int32 tensor of shape [batch_size].
      logits: A float32 tensor of shape [batch_size].
      alpha: A scalar for focal loss alpha hyper-parameter. If positive samples number
      < negative samples number, alpha < 0.5 and vice versa.
      gamma: A scalar for focal loss gamma hyper-parameter.
    Returns:
      A tensor of the same shape as `lables`
    """
    # alpha: the weight(unimportant) of negative
    # positive lable 1   * 1-alpha  0.75  loss更大了，更需要优化
    # negative lable 0   * alpha    0.25  loss更低了，更不需要优化
    y_pred = tf.nn.sigmoid(logits)
    labels = tf.to_float(labels)
    L = -labels * (1 - alpha) * ((1 - y_pred) ** gamma) * tf.log(y_pred) - \
        (1 - labels) * alpha * (y_pred ** gamma) * tf.log(1 - y_pred)
    return L


def focal_loss_softmax(labels, logits, gamma=2):
    """
    Computer focal loss for multi classification
    Args:
      labels: A int32 tensor of shape [batch_size].
      logits: A float32 tensor of shape [batch_size,num_classes].
      gamma: A scalar for focal loss gamma hyper-parameter.
    Returns:
      A tensor of the same shape as `lables`
    """
    y_pred = tf.nn.softmax(logits, dim=-1)  # [batch_size,num_classes]
    labels = tf.one_hot(labels, depth=y_pred.shape[-1])
    L = -labels * ((1 - y_pred) ** gamma) * tf.log(y_pred)
    L = tf.reduce_sum(L, axis=-1)
    return L


def cosine(vec1, vec2):
    # [-1, 1] greater is similar
    vec1_norm = tf.sqrt(tf.reduce_sum(tf.square(vec1), axis=-1))
    vec2_norm = tf.sqrt(tf.reduce_sum(tf.square(vec1), axis=-1))
    dot_product = tf.reduce_sum(tf.multiply(vec1, vec2), axis=-1)
    return dot_product / (vec1_norm * vec2_norm)


def dot_product(vec1, vec2, scale=False):
    # [-inf, inf] greater is similar
    dot = tf.reduce_sum(tf.multiply(vec1, vec2), axis=-1)
    if scale:
        dot = dot / tf.shpae(vec1)[-1] ** -0.5
    return dot


def euclidean(vec1, vec2):
    # [0, inf] smaller is similar
    return tf.sqrt(tf.reduce_sum(tf.square(vec1 - vec2), -1))


def l2_distance(vec1, vec2):
    # [0, inf] smaller is similar
    return tf.reduce_sum(tf.square(vec1 - vec2), -1)


def arccosine(vec1, vec2):
    # 余弦夹角
    # [0, 3.1415] smaller is similar
    return tf.math.acos(cosine(vec1, vec2))


def contrastive_loss(vec1, vec2, labels, metrics='euclidean', margin=1.25):
    # 对比损失 正样本对越相似 负样本对越不相似且达到阈值margin以上
    # vec [batch,hid]
    # labels [batch]
    # labels value:1/0  1 for pos, 0 for neg
    assert metrics in ['euclidean', 'arccosine', 'cosine', 'dot_product'], f'not support metrics {metrics}'

    labels = tf.cast(labels, tf.float32)

    if metrics in ['euclidean', 'arccosine']:
        if metrics == 'euclidean':
            distance = euclidean(vec1, vec2)  # [batch]
        else:
            distance = arccosine(vec1, vec2)  # [batch]
        loss = labels * tf.square(distance) + (1 - labels) * tf.square(tf.maximum(margin - distance), 0.)
        loss = tf.reduce_mean(loss, axis=0)

    if metrics in ['cosine', 'dot_product']:
        if metrics == 'cosine':
            similarity = cosine(vec1, vec2)  # [batch]
        else:
            similarity = dot_product(vec1, vec2, scale=True)  # [batch]
        loss = (1 - labels) * tf.square(similarity) + labels * tf.square(tf.maximum(margin - similarity), 0.)
        loss = tf.reduce_mean(loss, axis=0)

    return loss


def triplet_loss(anchor_vec, pos_vec, neg_vec, metrics='euclidean', margin=1.25):
    # vec [batch,hid]
    # 三元损失 正样本与锚样本相似要大于负样本与锚样本，且达到阈值margin以上
    assert metrics in ['euclidean', 'arccosine', 'cosine', 'dot_product'], f'not support metrics {metrics}'
    if metrics == 'euclidean':
        pos_pair = l2_distance(anchor_vec, pos_vec)
        neg_pair = l2_distance(anchor_vec, neg_vec)
    elif metrics == 'arccosine':
        pos_pair = arccosine(anchor_vec, pos_vec)
        neg_pair = arccosine(anchor_vec, neg_vec)
    elif metrics == 'cosine':
        pos_pair = cosine(anchor_vec, pos_vec)
        neg_pair = cosine(anchor_vec, neg_vec)
    else:
        pos_pair = dot_product(anchor_vec, pos_vec)
        neg_pair = dot_product(anchor_vec, neg_vec)

    if metrics in ['euclidean', 'arccosine']:
        loss = tf.maximum(pos_pair - neg_pair + margin, 0.)  # [batch]
    else:
        loss = tf.maximum(neg_pair - pos_pair + margin, 0.)  # [batch]
    loss = tf.reduce_mean(loss, axis=0)
    return loss


def label_smoothing(inputs, epsilon=0.1):
    '''Applies label smoothing. See https://arxiv.org/abs/1512.00567.

    Args:
      inputs: A 3d tensor with shape of [N, T, V], where V is the number of vocabulary.
      epsilon: Smoothing rate.

    For example,

    ```
    import tensorflow as tf
    inputs = tf.convert_to_tensor([[[0, 0, 1], 
       [0, 1, 0],
       [1, 0, 0]],

      [[1, 0, 0],
       [1, 0, 0],
       [0, 1, 0]]], tf.float32)

    outputs = label_smoothing(inputs)

    with tf.Session() as sess:
        print(sess.run([outputs]))

    >>
    [array([[[ 0.03333334,  0.03333334,  0.93333334],
        [ 0.03333334,  0.93333334,  0.03333334],
        [ 0.93333334,  0.03333334,  0.03333334]],

       [[ 0.93333334,  0.03333334,  0.03333334],
        [ 0.93333334,  0.03333334,  0.03333334],
        [ 0.03333334,  0.93333334,  0.03333334]]], dtype=float32)]   
    ```    
    '''
    K = inputs.get_shape().as_list()[-1]  # number of channels
    return ((1 - epsilon) * inputs) + (epsilon / K)


def center_loss_v3(features, labels, alpha, name='center_loss'):
    # features [batch,hid]
    # label one_hot [batch,num_class]
    hidden_size = features.get_shape().as_list()[-1]
    num_classes = labels.get_shape().as_list()[-1]
    labels = tf.argmax(labels, axis=-1)  # [batch]
    with tf.variable_scope(name):
        centers = tf.get_variable('centers', [num_classes, hidden_size],
                                  dtype=tf.float32,
                                  initializer=tf.contrib.layers.xavier_initializer(),
                                  trainable=False)
        centers_batch = tf.gather(centers, labels)  # 获取当前batch对应的类中心特征
        c_loss = tf.reduce_mean(tf.nn.l2_loss(features - centers_batch), axis=-1)

        centers_diff = alpha * (centers_batch - features)  # 类中心的梯度
        centers = tf.scatter_sub(centers, labels, centers_diff)  # 更新梯度
    return c_loss, centers


def softplus(x, name="softplus"):
    with tf.variable_scope(name):
        return tf.nn.softplus


def swish(x, name="swish"):
    """f(x) = sigmoid(x) * x
    """
    with tf.variable_scope(name):
        return (tf.nn.sigmoid(x * 1.0) * x)


def leaky_relu(x, leak=0.2, name="leaky_relu"):
    """f(x) = max(alpha * x, x)
    """
    with tf.variable_scope(name):
        f1 = 0.5 * (1 + leak)
        f2 = 0.5 * (1 - leak)
    return f1 * x + f2 * tf.abs(x)


def cube(x, name="cube_act"):
    """f(x) = pow(x, 3)
    """
    with tf.variable_scope(name):
        return tf.pow(x, 3)


def penalized_tanh(x, name="penalized_tanh"):
    """f(x) = max(tanh(x), alpha * tanh(x))
    """
    with tf.variable_scope(name):
        alpha = 0.25
        return tf.maximum(tf.tanh(x), alpha * tf.tanh(x))


def cosper(x, name="cosper_act"):
    """f(x) = cos(x) - x
    """
    with tf.variable_scope(name):
        return (tf.cos(x) - x)


def minsin(x, name="minsin_act"):
    """f(x) = min(x, xin(x))
    """
    with tf.variable_scope(name):
        return tf.minimum(x, tf.sin(x))


def tanhrev(x, name="tanhprev"):
    """f(x) = pow(atan(x), 2) - x
    """
    with tf.variable_scope(name):
        return (tf.pow(tf.atan(x), 2) - x)


def maxsig(x, name="maxsig_act"):
    """f(x) = max(x, tf.sigmiod(x))
    """
    with tf.variable_scope(name):
        return tf.maximum(x, tf.sigmoid(x))


def maxtanh(x, name="max_tanh_act"):
    """f(x) = max(x, tanh(x))
    """
    with tf.variable_scope(name):
        return tf.maximum(x, tf.tanh(x))


def get_activation(active_type='swish', **kwargs):
    mp = {
        "sigmoid": tf.nn.sigmoid,
        "tanh": tf.nn.tanh,
        "softsign": tf.nn.softsign,
        "relu": tf.nn.relu,
        "leaky_relu": leaky_relu,
        "elu": tf.nn.elu,
        "selu": tf.nn.selu,
        "swish": swish,
        "sin": tf.sin,
        "cube": cube,
        "penalized_tanh": penalized_tanh,
        "cosper": cosper,
        "minsin": minsin,
        "tanhrev": tanhrev,
        "maxsig": maxsig,
        "maxtanh": maxtanh,
        "softplus": tf.nn.softplus,
    }
    assert active_type in mp, "%s is not in activation list"
    return mp[active_type]


def cosine_proximity_tf(y_true, y_pred):
    """cosine loss"""
    y_true = tf.nn.l2_normalize(y_true, axis=-1)
    y_pred = tf.nn.l2_normalize(y_pred, axis=-1)
    return -(tf.reduce_sum(y_true * y_pred, axis=-1) + 1) / 2 + 1


def get_state_shape_invariants(tensor):
    """ return the shape of tensor but set middle dims to None """
    shape = tensor.shape.as_list()
    for i in range(1, len(shape) - 1):
        shape[i] = None
    return tf.TensorShape(shape)



#!/usr/bin/env python
# coding=utf-8
import os, re, glob
import numpy as np
import tensorflow as tf

""" tfrecord文件保存的名字后面均补上data数量
    e.g. xxx.tfrecord_1234
"""

TF_VERSION = int(tf.__version__.split('.')[1])


def tf_sparse_to_dense_new(v):
    if isinstance(v, tf.sparse.SparseTensor):
        return tf.sparse.to_dense(v)
    return v


def tf_sparse_to_dense_old(v):
    if isinstance(v, tf.SparseTensor):
        return tf.sparse_to_dense(v.indices, v.dense_shape, v.values)
    return v


tf_sparse_to_densor = tf_sparse_to_dense_new if TF_VERSION >= 12 else tf_sparse_to_dense_old


def flat(l):
    # 平摊flatten
    for k in l:
        if not isinstance(k, (list, tuple)):
            yield k
        else:
            yield from flat(k)


def exist_tfrecord_file(tfrecord_file):
    files = glob.glob(f'{tfrecord_file}_*')
    files = list(filter(lambda f: re.match(r'^.*_\d+$', f), files))  # filter invalid
    return True if len(files) > 0 else False


def get_tfrecord_file(tfrecord_file):
    # 获得对应包括num的tfrecord名字
    files = glob.glob(f'{tfrecord_file}_*')
    files = list(filter(lambda f: re.match(r'^.*_\d+$', f), files))  # filter invalid
    if len(files) > 0:
        if len(files) > 1:
            files.sort(key=lambda f: os.path.getctime(f), reverse=True)  # 优先拿最新的
        return files[0]
    else:
        return None


def delete_exist_tfrecord_file(tfrecord_file):
    files = glob.glob(f'{tfrecord_file}_*')
    files = list(filter(lambda f: re.match(r'^.*_\d+$', f), files))  # filter invalid
    for f in files:
        if os.path.isfile(f):
            os.remove(f)


def add_num(file, num):
    return f'{file}_{num}'


def items2tfrecord(items, tfrecord_file):
    # 删除已有的
    delete_exist_tfrecord_file(tfrecord_file)

    def int_feat(value):
        if not isinstance(value, (list, np.ndarray)):
            value = [value]
        return tf.train.Feature(int64_list=tf.train.Int64List(value=value))

    def float_feat(value):
        if not isinstance(value, (list, np.ndarray)):
            value = [value]
        return tf.train.Feature(float_list=tf.train.FloatList(value=value))

    def byte_feat(value):
        if not isinstance(value, (list, np.ndarray)):
            value = [value]
        return tf.train.Feature(bytesList=tf.train.BytesList(value=value))

    writer = tf.python_io.TFRecordWriter(tfrecord_file)
    count = 0
    for item in items:
        if count and not count % 100000:
            print(f'generating tfrecord... count: {count}')
        features = {}
        try:
            for k, v in item.items():
                # maybe_flatten
                if isinstance(v, np.ndarray) and v.ndim > 1:
                    v = v.flatten()
                elif isinstance(v, list) and isinstance(v[0], list):
                    v = list(flat(v))

                ele = v[0] if isinstance(v, (list, np.ndarray)) else v  # 需检查元素的类型

                if isinstance(ele, (int, np.int, np.int32, np.int64)):
                    features[k] = int_feat(v)
                elif isinstance(ele, (float, np.float, np.float16, np.float32, np.float64)):
                    features[k] = float_feat(v)
                else:
                    features[k] = byte_feat(v)
            example = tf.train.Example(features=tf.train.Features(feature=features))
        except:
            print('error item:', item)
            continue
        writer.write(example.SerializeToString())
        count += 1
    writer.close()
    add_num_name = add_num(tfrecord_file, count)
    os.rename(tfrecord_file, add_num_name)
    if count == 0:
        raise Exception(f'error! count = {count} no example to save')
    print(f'save tfrecord file ok! {add_num_name} total count: {count}')
    return count


def tfrecord2dataset(tfrecord_files, feat_dct, shape_dct=None, batch_size=100, auto_pad=False, index=None, shard=None):
    """
    tf.VarLenFeature只能是平展后的1维数据，故原始数据是二维的话，如[None,4]，则需传入shape_dct进行reshape
    feat_dct = {
       'target': tf.FixedLenFeature([], tf.int64),
       's1': tf.VarLenFeature(tf.int64),
       's1_char': tf.VarLenFeature(tf.int64),
       'others': tf.FixedLenFeature([3], tf.string),
    }
    shape_dct = {
       's1_char': [-1, 4]  # if s1_char is 2-D Tensor and need to pad at the first dimension when batch
    }
    """
    # [注]该函数返回dataset的过程不需在图中,但后续迭代需在图中使用,如:
    # with self.graph.as_default():
    #     iterator = dataset.make_one_shot_iterator()
    #     dataset_features = iterator.get_next()
    # for i in train_step:
    #     features = sess.run(dataset_features)

    # 首先根据传入的tfrecord_file名字得到num
    if not isinstance(tfrecord_files, list):
        tfrecord_files = [tfrecord_files]
    total_count = 0
    files = []
    for file in tfrecord_files:
        if re.match(r'^.*_\d+$', file):  # 传入的是已经有数字了
            files.append(file)
            total_count += int(file.rsplit('_', 1)[1])
            continue
        file_ = get_tfrecord_file(file)
        if file_ is None:
            print(f'valid tfrecord(with num) file {file} is not found!')
            continue
        files.append(file_)
        total_count += int(file_.rsplit('_', 1)[1])
    tfrecord_files = files
    print(f'load tfrecord file ok! {" & ".join(tfrecord_files)} total count: {total_count}')

    def exm_parse(serialized_example):
        parsed_features = tf.parse_single_example(serialized_example, features=feat_dct)
        # VarLenFeature will return sparse tensor
        for k, v in parsed_features.items():
            parsed_features[k] = tf_sparse_to_densor(v)  # Convert sparse to dense when need
        # maybe need to reshape
        if shape_dct is not None:
            for name, shape in shape_dct.items():
                parsed_features[name] = tf.reshape(parsed_features[name], shape)
        return parsed_features

    def ints2int32(features):
        for k, v in features.items():
            if v.dtype in [tf.int64, tf.uint8]:
                features[k] = tf.cast(v, tf.int32)
        return features

    def padded_batch(dataset, batch_size, padded_shapes=None):
        if padded_shapes is None:
            padded_shapes = {k: [None] * len(shape) for k, shape in dataset.output_shapes.items()}
        return dataset.padded_batch(batch_size, padded_shapes)

    # tf.contrib.keras.preprocessing.sequence.pad_sequences

    dataset = tf.data.TFRecordDataset(tf.constant(tfrecord_files))
    # dataset = tf.data.TFRecordDataset(tf.data.Dataset.from_tensor_slices(tf.constant(data_files)))

    # 是否分布式训练数据分片, 上层传来的local_rank设置shard的数据，保证各个gpu采样的数据不重叠。
    if shard is not None and index is not None and index < shard:
        dataset = dataset.shard(shard, index)
        batch_size = int(np.ceil(batch_size / shard))  # batch再均分
        

    dataset = dataset.map(exm_parse)
    # dataset = dataset.map(exm_parse, num_parallel_calls=tf.data.experimental.AUTOTUNE)  # need tf >= 1.14

    dataset = dataset.shuffle(buffer_size=5000, reshuffle_each_iteration=True)

    dataset = padded_batch(dataset, batch_size) if auto_pad else dataset.batch(batch_size)  # batch时是否要自动补齐

    dataset = dataset.map(ints2int32)  # 可向量化的操作放在batch后面以提高效率
    # dataset = dataset.map(ints2int32, num_parallel_calls=tf.data.experimental.AUTOTUNE)  # 可向量化的操作放在batch后面以提高效率

    dataset = dataset.repeat()  # repeat放在batch后面

    dataset = dataset.prefetch(2)  # pipline先异步准备好n个batch数据,训练step时数据同时处理
    # dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)  # pipline先异步准备好n个batch数据,训练step时数据同时处理

    return dataset, total_count


def tfrecord2queue(tfrecord_files, feat_dct):
    # 返回后还需使用
    # coord = tf.train.Coordinator()
    # threads = tf.train.start_queue_runners(coord=coord)
    # sess.run(tfrecord2queue)
    # coord.request_stop()
    # coord.join(threads)
    filename_queue = tf.train.string_input_producer(tf.train.match_filenames_once(tfrecord_files),
                                                    shuffle=None,
                                                    num_epochs=None)
    reader = tf.TFRecordReader()
    _, serialized_example = reader.read(filename_queue)

    features = tf.parse_single_example(serialized_example, features=feat_dct)

    for k in features:
        if features[k].dtype in [tf.int64, tf.uint8]:
            features[k] = tf.cast(features[k], tf.int32)

    # reshpae
    # features[k] = tf.reshape(features[k], [-1,1])
    sorted_keys = list(sorted(features))

    input_queue = tf.train.slice_input_producer([features[k] for k in sorted_keys], shuffle=False)
    # tf.data.Dataset.from_tensor_slices
    data = tf.train.batch(input_queue, batch_size=128, allow_smaller_final_batch=True, num_threads=8)
    ret = dict(zip(sorted_keys, data))
    return ret  # 返回的tensor供sess.run


def check_tfrecord(tfrecord_file, feat_dct):
    true_count = 0
    dataset, total_count = tfrecord2dataset(tfrecord_file, feat_dct)
    iterator = dataset.make_one_shot_iterator()
    features = iterator.get_next()
    sess = tf.Session()
    while True:
        try:
            feat = sess.run(features)
            true_count += len(list(feat.values())[0])
        except Exception as e:
            print(e)
            print(true_count)
            raise e


if __name__ == '__main__':
    # os.environ['CUDA_VISIBLE_DEVICES'] = '6'
    #
    # feat_dct = {
    #     'contents': tf.FixedLenFeature([4, 50], tf.int64),
    #     'content_masks': tf.FixedLenFeature([4, 50], tf.int64),
    #     'content_lens': tf.FixedLenFeature([4], tf.int64),
    #     'char_contents': tf.FixedLenFeature([4, 50, 4], tf.int64),
    #     'char_content_masks': tf.FixedLenFeature([4, 50, 4], tf.int64),
    #     'char_content_lens': tf.FixedLenFeature([4, 50], tf.int64),
    #     'responses': tf.FixedLenFeature([50], tf.int64),
    #     'response_masks': tf.FixedLenFeature([50], tf.int64),
    #     'response_lens': tf.FixedLenFeature([], tf.int64),
    #     'char_responses': tf.FixedLenFeature([50, 4], tf.int64),
    #     'char_response_masks': tf.FixedLenFeature([50, 4], tf.int64),
    #     'char_response_lens': tf.FixedLenFeature([50], tf.int64),
    #     'targets': tf.FixedLenFeature([], tf.int64),
    #     'intents': tf.FixedLenFeature([4], tf.int64),
    # }
    # check_tfrecord('/dockerdata/yonaszhang/1223/data/match_ckpt_4_train_data.tfrecord_1', feat_dct)
    # exit(0)

    """ test """
    import time

    # items = [
    #     {'s1':[1,2,3,4],'s2':[5]},
    #     {'s1':[2,3,4],'s2':[6]},
    # ]
    # items2tfrecord(items, './test.tfrecord')
    feat_dct = {
        's1': tf.VarLenFeature(tf.int64),
        's2': tf.FixedLenFeature([], tf.int64),
    }
    dataset, _ = tfrecord2dataset('./test.tfrecord', feat_dct, batch_size=2, auto_pad=True)
    feature = dataset.make_one_shot_iterator().get_next()
    with tf.Session() as sess:
        while True:
            print(sess.run(feature))
            time.sleep(1)
            # input('输入任意键以继续下一个')



import os
import tensorflow as tf
import numpy as np

curr_dir = os.path.dirname(os.path.realpath(__file__))

from qiznlp.common.modules.common_layers import mask_nonpad_from_embedding
from qiznlp.common.modules.embedding import embedding
from qiznlp.common.modules.birnn import Bi_RNN
import qiznlp.common.utils as utils

conf = utils.dict2obj({
    "vocab_size": 17885,
    "embed_size": 300,
    'birnn_hidden_size': 300,
    'l2_reg': 0.0001,
    'dropout_rate': 0.2,
    'lr': 1e-3,
    'pretrain_emb': None,
})


class Model(object):
    def __init__(self, build_graph=True, **kwargs):
        self.conf = conf
        if build_graph:
            # build placeholder
            self.build_placeholder()
            # build model
            self.model_name = kwargs.get('model_name', 'esim')
            {
                'esim': self.build_model,
                # add new here
            }[self.model_name]()
            print(f'model_name: {self.model_name} build graph ok!')

    def build_placeholder(self):
        # placeholder
        # 原则上模型输入输出不变，不需换新model
        self.s1 = tf.placeholder(tf.int32, [None, None], name='s1')
        self.s2 = tf.placeholder(tf.int32, [None, None], name='s2')
        self.target = tf.placeholder(tf.int32, [None], name="target")
        self.dropout_rate = tf.placeholder(tf.float32, name="dropout_rate")

    def build_model(self):
        # embedding
        # [batch,len,embed]
        s1_embed, _ = embedding(tf.expand_dims(self.s1, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)
        s2_embed, _ = embedding(tf.expand_dims(self.s2, -1), conf.vocab_size, conf.embed_size, name='share_embedding', pretrain_embedding=conf.pretrain_emb)

        s1_input_mask = mask_nonpad_from_embedding(s1_embed)  # [batch,len1] 1 for nonpad; 0 for pad
        s2_input_mask = mask_nonpad_from_embedding(s2_embed)  # [batch,len2] 1 for nonpad; 0 for pad
        s1_seq_len = tf.cast(tf.reduce_sum(s1_input_mask, axis=-1), tf.int32)  # [batch]
        s2_seq_len = tf.cast(tf.reduce_sum(s2_input_mask, axis=-1), tf.int32)  # [batch]

        # bilstm sent encoder
        self.bilstm_encoder1 = Bi_RNN(cell_name='LSTMCell', hidden_size=conf.birnn_hidden_size, dropout_rate=self.dropout_rate)
        s1_bar, _ = self.bilstm_encoder1(s1_embed, s1_seq_len)  # [batch,len1,2hid]
        s2_bar, _ = self.bilstm_encoder1(s2_embed, s2_seq_len)  # [batch,len2,2hid]

        # local inference 局部推理
        with tf.variable_scope('local_inference'):
            # 点积注意力
            attention_logits = tf.matmul(s1_bar, tf.transpose(s2_bar, [0, 2, 1]))  # [batch,len1,len2]

            # 注意需attention mask  pad_mask * -inf + logits
            attention_s1 = tf.nn.softmax(attention_logits + tf.expand_dims((1. - s2_input_mask) * -1e9, 1))  # [batch,len1,len2]

            attention_s2 = tf.nn.softmax(tf.transpose(attention_logits, [0, 2, 1]) + tf.expand_dims((1. - s1_input_mask) * -1e9, 1))  # [batch,len2,len1]

            s1_hat = tf.matmul(attention_s1, s2_bar)  # [batch,len1,2hid]
            s2_hat = tf.matmul(attention_s2, s1_bar)  # [batch,len2,2hid]

            s1_diff = s1_bar - s1_hat
            s1_mul = s1_bar * s1_hat

            s2_diff = s2_bar - s2_hat
            s2_mul = s2_bar * s2_hat

            m_s1 = tf.concat([s1_bar, s1_hat, s1_diff, s1_mul], axis=2)  # [batch,len1,8hid]
            m_s2 = tf.concat([s2_bar, s2_hat, s2_diff, s2_mul], axis=2)  # [batch,len2,8hid]

        # composition 推理组成
        with tf.variable_scope('composition'):
            self.bilstm_encoder2 = Bi_RNN(cell_name='LSTMCell', hidden_size=conf.birnn_hidden_size, dropout_rate=self.dropout_rate)
            v_s1, _ = self.bilstm_encoder2(m_s1, s1_seq_len)  # [batch,len1,2hid]
            v_s2, _ = self.bilstm_encoder2(m_s2, s2_seq_len)  # [batch,len2,2hid]

            # average pooling  # 需将pad的vector变为0
            v_s1 = v_s1 * tf.expand_dims(s1_input_mask, -1)  # [batch,len1,2hid]
            v_s2 = v_s2 * tf.expand_dims(s2_input_mask, -1)  # [batch,len1,2hid]
            v_s1_avg = tf.reduce_sum(v_s1, axis=1) / tf.cast(tf.expand_dims(s1_seq_len, -1), tf.float32)  # [batch,2hid]
            v_s2_avg = tf.reduce_sum(v_s2, axis=1) / tf.cast(tf.expand_dims(s2_seq_len, -1), tf.float32)  # [batch,2hid]

            # max pooling  # 需将pad的vector变为极小值
            v_s1_max = tf.reduce_max(v_s1 + tf.expand_dims((1. - s1_input_mask) * -1e9, -1), axis=1)  # [batch,2hid]
            v_s2_max = tf.reduce_max(v_s2 + tf.expand_dims((1. - s2_input_mask) * -1e9, -1), axis=1)  # [batch,2hid]

            v = tf.concat([v_s1_avg, v_s1_max, v_s2_avg, v_s2_max], axis=-1)  # [batch,8hid]

        with tf.variable_scope('ffn'):
            h_ = tf.layers.dropout(v, rate=self.dropout_rate)
            h = tf.layers.dense(h_, 256, activation=tf.nn.relu, kernel_initializer=tf.random_normal_initializer(0.0, 0.1))  # [batch,256]
            o_ = tf.layers.dropout(h, rate=self.dropout_rate)
            o = tf.layers.dense(o_, 1, kernel_initializer=tf.random_normal_initializer(0.0, 0.1))  # [batch,1]
        self.logits = tf.squeeze(o, -1)  # [batch]

        # loss
        with tf.name_scope('loss'):
            loss = tf.nn.sigmoid_cross_entropy_with_logits(labels=tf.cast(self.target, tf.float32), logits=self.logits)  # [batch]
            loss = tf.reduce_mean(loss, -1)  # scalar

            l2_reg = conf.l2_reg
            weights = [v for v in tf.trainable_variables() if ('w' in v.name) or ('kernel' in v.name)]
            l2_loss = tf.add_n([tf.nn.l2_loss(w) for w in weights]) * l2_reg
            loss += l2_loss

        self.loss = loss
        self.y_prob = tf.nn.sigmoid(self.logits)
        self.y_prob = tf.identity(self.y_prob, name='y_prob')

        with tf.name_scope("accuracy"):
            self.correct = tf.equal(
                tf.cast(tf.greater_equal(self.y_prob, 0.5), tf.int32),
                self.target)
            self.accuracy = tf.reduce_mean(tf.cast(self.correct, tf.float32))

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    @classmethod
    def sent2ids(cls, sent, word2id, max_word_len=None):
        # sent 已分好词 ' '隔开
        # 形成batch时才动态补齐长度
        words = sent.split(' ')
        token_ids = [word2id.get(word, word2id['<unk>']) for word in words]
        if max_word_len:
            token_ids = token_ids[:max_word_len]
        # token_ids = utils.pad_sequences([token_ids], padding='post', maxlen=max_word_len)[0]
        return token_ids  # [len]

    def create_feed_dict_from_data(self, data, ids, mode='train'):
        # data:数据已经转为id, data不同字段保存该段字段全量数据
        batch_s1 = [data['s1'][i] for i in ids]
        batch_s2 = [data['s2'][i] for i in ids]
        if len(set([len(e) for e in batch_s1])) != 1:  # 长度不等
            batch_s1 = utils.pad_sequences(batch_s1, padding='post')
        if len(set([len(e) for e in batch_s2])) != 1:  # 长度不等
            batch_s2 = utils.pad_sequences(batch_s2, padding='post')
        feed_dict = {
            self.s1: batch_s1,
            self.s2: batch_s2,
            self.target: [data['target'][i] for i in ids],
        }
        if mode == 'train': feed_dict['num'] = len(batch_s1)
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_features(self, features, mode='train'):
        # feature:tfrecord数据的example, 每个features的不同字段包括该字段一个batch数据
        feed_dict = {
            self.s1: features['s1'],
            self.s2: features['s2'],
            self.target: features['target'],
        }
        if mode == 'train': feed_dict['num'] = len(features['s1'])
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_raw(self, batch_s1, batch_s2, batch_y, token2id_dct, mode='infer'):
        word2id = token2id_dct['word2id']
        
        feed_s1 = [self.sent2ids(s1, word2id) for s1 in batch_s1]
        if len(set([len(e) for e in feed_s1])) != 1:  # 长度不等
            feed_s1 = utils.pad_sequences(feed_s1, padding='post')

        feed_s2 = [self.sent2ids(s2, word2id) for s2 in batch_s2]
        if len(set([len(e) for e in feed_s2])) != 1:  # 长度不等
            feed_s2 = utils.pad_sequences(feed_s2, padding='post')

        feed_dict = {
            self.s1: np.array(feed_s1),
            self.s2: np.array(feed_s2),
        }
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.

        if mode == 'infer':
            return feed_dict

        if mode in ['train', 'dev']:
            feed_dict[self.target] = batch_y
            return feed_dict
        
        raise ValueError(f'mode type {mode} not support')

    @classmethod
    def generate_data(cls, file, token2id_dct):
        word2id = token2id_dct['word2id']
        data = {
            's1': [],
            's2': [],
            'target': []
        }
        with open(file, 'r', encoding='U8') as f:
            for i, line in enumerate(f):
                item = line.strip().split('\t')
                if len(item) != 3:
                    print('error', repr(line))
                    continue
                s1 = item[0]
                s2 = item[1]
                y = item[2]
                s1_ids = cls.sent2ids(s1, word2id, max_word_len=50)
                s2_ids = cls.sent2ids(s2, word2id, max_word_len=50)
                y_id = int(y)
                if i < 5:  # check
                    print(f'check {i}:')
                    print(f'{s1} -> {s1_ids}')
                    print(f'{s2} -> {s2_ids}')
                    print(f'{y} -> {y_id}')
                data['s1'].append(s1_ids)
                data['s2'].append(s2_ids)
                data['target'].append(y_id)
        data['num_data'] = len(data['s1'])
        return data

    @classmethod
    def generate_tfrecord(cls, file, token2id_dct, tfrecord_file):
        from qiznlp.common.tfrecord_utils import items2tfrecord
        word2id = token2id_dct['word2id']

        def items_gen():
            with open(file, 'r', encoding='U8') as f:
                for i, line in enumerate(f):
                    item = line.strip().split('\t')
                    if len(item) != 3:
                        print('error', repr(line))
                        continue
                    try:
                        s1 = item[0]
                        s2 = item[1]
                        y = item[2]
                        s1_ids = cls.sent2ids(s1, word2id, max_word_len=50)
                        s2_ids = cls.sent2ids(s2, word2id, max_word_len=50)
                        y_id = int(y)
                        if i < 5:  # check
                            print(f'check {i}:')
                            print(f'{s1} -> {s1_ids}')
                            print(f'{s2} -> {s2_ids}')
                            print(f'{y} -> {y_id}')
                        d = {
                            's1': s1_ids,
                            's2': s2_ids,
                            'target': y_id,
                        }
                        yield d
                    except:
                        continue

        count = items2tfrecord(items_gen(), tfrecord_file)
        return count

    @classmethod
    def load_tfrecord(cls, tfrecord_file, batch_size=128, index=None, shard=None):
        from qiznlp.common.tfrecord_utils import tfrecord2dataset, exist_tfrecord_file
        if not exist_tfrecord_file(tfrecord_file):
            return None, None
        feat_dct = {
            # 's1': tf.FixedLenFeature([50], tf.int64),
            's1': tf.VarLenFeature(tf.int64),
            's2': tf.VarLenFeature(tf.int64),
            'target': tf.FixedLenFeature([], tf.int64),
        }
        dataset, count = tfrecord2dataset(tfrecord_file, feat_dct, batch_size=batch_size, auto_pad=True, index=index, shard=shard)
        return dataset, count

    def get_signature_export_model(self):
        inputs_dct = {
            's1': self.s1,
            's2': self.s2,
            'dropout_rate': self.dropout_rate,
        }
        outputs_dct = {
            'y_prob': self.y_prob,
        }
        return inputs_dct, outputs_dct

    @classmethod
    def get_signature_load_pbmodel(cls):
        inputs_lst = ['s1', 's2', 'dropout_rate']
        outputs_lst = ['y_prob']
        return inputs_lst, outputs_lst

    @classmethod
    def from_pbmodel(cls, pbmodel_dir, sess):
        meta_graph_def = tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], pbmodel_dir)  # 从pb模型载入graph和variable,绑定到sess
        signature = meta_graph_def.signature_def  # 取出signature_def
        default_key = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY  # or 'chitchat_predict' 签名
        inputs_lst, output_lst = cls.get_signature_load_pbmodel()  # 类方法
        pb_dict = {}
        for k in inputs_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].inputs[k].name)  # 从signature中获取输入输出的tensor_name,并从graph中取出
        for k in output_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].outputs[k].name)
        model = cls(build_graph=False)  # 里面不再构造图
        for k, v in pb_dict.items():
            setattr(model, k, v)  # 绑定必要的输入输出到实例
        return model

    @classmethod
    def from_ckpt_meta(cls, ckpt_name, sess, graph):
        with graph.as_default():
            saver = tf.train.import_meta_graph(ckpt_name + '.meta', clear_devices=True)
            sess.run(tf.global_variables_initializer())

        model = cls(build_graph=False)  # 里面不再构造图
        # 绑定必要的输入输出到实例
        model.s1 = graph.get_tensor_by_name('s1:0')
        model.s2 = graph.get_tensor_by_name('s2:0')
        model.dropout_rate = graph.get_tensor_by_name('dropout_rate:0')
        # self.target = self.graph.get_tensor_by_name('target:0')
        model.y_prob = graph.get_tensor_by_name('y_prob:0')

        saver.restore(sess, ckpt_name)
        print(f':: restore success! {ckpt_name}')
        return model, saver



#!/usr/bin/env python
# coding: utf-8
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='QizNLP',
    version='0.1.0',
    author='Qznan',
    author_email='summerzynqz@gmail.com',
    description='Quick run NLP in many task',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Qznan/QizNLP',
    packages=find_packages(),
    package_data={'qiznlp': ['data/*.txt'], },
    install_requires=[
        'jieba',
        'tensorflow>=1.8, <=1.12'
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'qiznlp_init=qiznlp:qiznlp_init',
        ]
    }
)



#!/usr/bin/env python
# coding=utf-8
from .common_layers import *


def embedding(ids, vocab_size, embedding_size, name='embedding', reuse=tf.AUTO_REUSE, pad_id=0, scale_sqrt_depth=True,
              pretrain_embedding=None, pretrain_trainable=True, word_dropout_rate=0.):
    """ embedding """
    # ids 3-D Tensor [batch, length, 1]
    if pretrain_embedding is None:
        with tf.variable_scope(name, reuse=reuse):
            var = tf.get_variable('weights', [vocab_size, embedding_size],  # [vocab,embed]
                                  initializer=tf.random_normal_initializer(0.0, embedding_size ** -0.5))
    else:
        with tf.variable_scope(name, reuse=reuse):
            var = tf.get_variable('weights', [vocab_size, embedding_size],  # [vocab,embed]
                                  trainable=pretrain_trainable,
                                  initializer=tf.constant_initializer(pretrain_embedding, dtype=tf.float32))

    # word level drop out
    if word_dropout_rate:
        ids = dropout_no_scaling(ids, 1.0 - word_dropout_rate)  # 随机将部分id变为0,相当于将单词变为pad

    # lookup table
    embedding = tf.gather(var, ids)  # [batch,length,1,hidden]
    embedding = tf.squeeze(embedding, axis=-2)  # [batch,length,hidden]
    if scale_sqrt_depth:
        embedding *= embedding_size ** 0.5
    embedding = embedding * tf.to_float(tf.not_equal(ids, pad_id))  # 将pad(id=0)的emb变为[0,0,...]
    return embedding, var


def proj_logits(outputs, hidden_size, logit_size, name='proj_logits', reuse=tf.AUTO_REUSE):
    """ if name = 'embedding' 复用embed矩阵 
        outputs [batch, length, hidden] or [batch, hidden]
    """

    with tf.variable_scope(name, reuse=reuse):
        var = tf.get_variable('weights', [logit_size, hidden_size],  # [vocab,hidden]
                              initializer=tf.random_normal_initializer(0.0, hidden_size ** -0.5))

    outputs_shape = shape_list(outputs)  # [batch, length, hidden]
    outputs = tf.reshape(outputs, [-1, outputs_shape[-1]])  # [batch*length,hidden]
    logits = tf.matmul(outputs, var, transpose_b=True)  # x,h * h,l -> x,l
    logits = tf.reshape(logits, outputs_shape[:-1] + [logit_size])  # [batch,length,vocab]

    return logits



import os
import tensorflow as tf
import numpy as np

curr_dir = os.path.dirname(os.path.realpath(__file__))

from qiznlp.common.modules.common_layers import mask_nonpad_from_embedding, add_timing_signal_1d, label_smoothing, softmax_cross_entropy
from qiznlp.common.modules.embedding import embedding
from qiznlp.common.modules.encoder import transformer_encoder, mean_pool, multi_head_attention_base_pool
import qiznlp.common.utils as utils

conf = utils.dict2obj({
    "vocab_size": 14180,
    'label_size': 16,
    "embed_size": 300,
    'hidden_size': 300,
    'num_heads': 6,
    'num_encoder_layers': 6,
    'dropout_rate': 0.2,
    'lr': 1e-3,
    'pretrain_emb': None,
    # 'pretrain_emb': np.load(f'{curr_dir}/pretrain_word_emb300.npy'),
})


class Model(object):
    def __init__(self, build_graph=True, **kwargs):

        self.conf = conf
        if build_graph:
            # build placeholder
            self.build_placeholder()
            # build model
            self.model_name = kwargs.get('model_name', 'trans_meanpool')
            {
                'trans_meanpool': self.build_model,
                'trans_mhattnpool': self.build_model1,
                # add new here
            }[self.model_name]()
            print(f'model_name: {self.model_name} build graph ok!')

    def build_placeholder(self):
        # placeholder
        # 原则上模型输入输出不变，不需换新model
        self.s1 = tf.placeholder(tf.int32, [None, None], name='s1')
        self.target = tf.placeholder(tf.int32, [None], name="target")
        self.dropout_rate = tf.placeholder(tf.float32, name="dropout_rate")

    def build_model(self):
        s1_embed, _ = embedding(tf.expand_dims(self.s1, -1), conf.vocab_size, conf.embed_size, pretrain_embedding=conf.pretrain_emb)

        """ encoder """
        # encoder_input  # [batch,length,embed]
        encoder_input = s1_embed
        encoder_valid_mask = mask_nonpad_from_embedding(encoder_input)  # [batch,length] 1 for nonpad; 0 for pad
        encoder_input = add_timing_signal_1d(encoder_input)  # add position embedding
        encoder_input = tf.layers.dropout(encoder_input, self.dropout_rate)  # dropout

        encoder_output = transformer_encoder(encoder_input, encoder_valid_mask,
                                             hidden_size=conf.hidden_size,
                                             filter_size=conf.hidden_size * 4,
                                             num_heads=conf.num_heads,
                                             num_encoder_layers=conf.num_encoder_layers,
                                             dropout=self.dropout_rate,
                                             attention_dropout=self.dropout_rate,
                                             relu_dropout=self.dropout_rate,
                                             )

        mean_pool_output = mean_pool(encoder_output, encoder_valid_mask)

        self.logits = tf.layers.dense(mean_pool_output, conf.label_size, use_bias=True, name='logits')  # [batch,cls]
        self.one_hot_target = tf.one_hot(self.target, conf.label_size)
        smooth_one_hot_target = label_smoothing(self.one_hot_target)
        self.loss = softmax_cross_entropy(self.logits, smooth_one_hot_target)
        self.y_prob = tf.nn.softmax(self.logits)
        self.y_prob = tf.identity(self.y_prob, name='y_prob')

        with tf.name_scope("accuracy"):
            self.correct = tf.equal(
                tf.argmax(self.logits, -1, output_type=tf.int32),
                self.target)
            self.accuracy = tf.reduce_mean(tf.cast(self.correct, tf.float32))

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    def build_model1(self):
        # pretrained_word_embeddings = np.load(f'{curr_dir}/pretrain_emb_300.npy')
        pretrained_word_embeddings = None
        s1_embed, _ = embedding(tf.expand_dims(self.s1, -1), conf.vocab_size, conf.embed_size, pretrain_embedding=pretrained_word_embeddings)

        """ encoder """
        # encoder_input  # [batch,length,embed]
        encoder_input = s1_embed
        encoder_valid_mask = mask_nonpad_from_embedding(encoder_input)  # [batch,length] 1 for nonpad; 0 for pad
        encoder_input = add_timing_signal_1d(encoder_input)  # add position embedding
        encoder_input = tf.layers.dropout(encoder_input, self.dropout_rate)  # dropout

        encoder_output = transformer_encoder(encoder_input, encoder_valid_mask,
                                             hidden_size=conf.hidden_size,
                                             filter_size=conf.hidden_size * 4,
                                             num_heads=conf.num_heads,
                                             num_encoder_layers=conf.num_encoder_layers,
                                             dropout=self.dropout_rate,
                                             attention_dropout=self.dropout_rate,
                                             relu_dropout=self.dropout_rate,
                                             )

        # attention # e.g. HAN
        attn_output = multi_head_attention_base_pool(encoder_output, encoder_valid_mask,
                                                     conf.embed_size,
                                                     conf.embed_size,
                                                     6,
                                                     name='sent_pool',
                                                     reuse=tf.AUTO_REUSE,
                                                     dropout=self.dropout_rate,
                                                     attn_v=None)

        self.logits = tf.layers.dense(attn_output, conf.label_size, use_bias=True, name='logits')  # [batch,cls]
        self.one_hot_target = tf.one_hot(self.target, conf.label_size)
        smooth_one_hot_target = label_smoothing(self.one_hot_target)
        self.loss = softmax_cross_entropy(self.logits, smooth_one_hot_target)
        self.y_prob = tf.nn.softmax(self.logits)
        self.y_prob = tf.identity(self.y_prob, name='y_prob')

        with tf.name_scope("accuracy"):
            self.correct = tf.equal(
                tf.argmax(self.logits, -1, output_type=tf.int32),
                self.target)
            self.accuracy = tf.reduce_mean(tf.cast(self.correct, tf.float32))

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    @classmethod
    def sent2ids(cls, sent, word2id, max_word_len=None):
        # sent 已分好词 ' '隔开
        # 形成batch时才动态补齐长度
        words = sent.split(' ')
        token_ids = [word2id.get(word, word2id['<unk>']) for word in words]
        return token_ids  # [len]

    @classmethod
    def label2id(cls, label, label2id):
        label_id = label2id.get(label, label2id['<unk>'])
        return label_id

    def create_feed_dict_from_data(self, data, ids, mode='train'):
        # data:数据已经转为id, data不同字段保存该段字段全量数据
        batch_s1 = [data['s1'][i] for i in ids]
        if len(set([len(e) for e in batch_s1])) != 1:  # 长度不等
            batch_s1 = utils.pad_sequences(batch_s1, padding='post')
        feed_dict = {
            self.s1: batch_s1,
            self.target: [data['target'][i] for i in ids],
        }
        if mode == 'train': feed_dict['num'] = len(batch_s1)
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_features(self, features, mode='train'):
        # feature:tfrecord数据的example, 每个features的不同字段包括该字段一个batch数据
        feed_dict = {
            self.s1: features['s1'],
            self.target: features['target'],
        }
        if mode == 'train': feed_dict['num'] = len(features['s1'])
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_raw(self, batch_s1, batch_y, token2id_dct, mode='infer'):
        word2id = token2id_dct['word2id']
        label2id = token2id_dct['label2id']

        feed_s1 = [self.sent2ids(s1, word2id) for s1 in batch_s1]
        if len(set([len(e) for e in feed_s1])) != 1:  # 长度不等
            feed_s1 = utils.pad_sequences(feed_s1, padding='post')

        feed_dict = {
            self.s1: np.array(feed_s1),
        }
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.

        if mode == 'infer':
            return feed_dict

        if mode in ['train', 'dev']:
            feed_target = [self.label2id(y, label2id) for y in batch_y]
            feed_dict[self.target] = feed_target
            return feed_dict
        raise ValueError(f'mode type {mode} not support')

    @classmethod
    def generate_data(cls, file, token2id_dct):
        word2id = token2id_dct['word2id']
        label2id = token2id_dct['label2id']
        data = {
            's1': [],
            'target': []
        }
        with open(file, 'r', encoding='U8') as f:
            for i, line in enumerate(f):
                item = line.strip().split('\t')
                s1 = item[0]
                y = item[1]
                s1_ids = cls.sent2ids(s1, word2id)
                s1_ids = s1_ids[:50]  # 长度截断
                y_id = cls.label2id(y, label2id)
                data['s1'].append(s1_ids)
                data['target'].append(y_id)
        data['num_data'] = len(data['s1'])
        return data

    @classmethod
    def generate_tfrecord(cls, file, token2id_dct, tfrecord_file):
        from qiznlp.common.tfrecord_utils import items2tfrecord
        word2id = token2id_dct['word2id']
        label2id = token2id_dct['label2id']

        def items_gen():
            with open(file, 'r', encoding='U8') as f:
                for i, line in enumerate(f):
                    item = line.strip().split('\t')
                    if len(item) != 2:
                        print(repr(line))
                        continue
                    try:
                        s1 = item[0]
                        y = item[1]
                        s1_ids = cls.sent2ids(s1, word2id)
                        y_id = cls.label2id(y, label2id)
                        s1_ids = s1_ids[:50]  # 截断
                        if i < 5:  # check
                            print(f'check {i}:')
                            print(f'{s1} -> {s1_ids}')
                            print(f'{y} -> {y_id}')
                        d = {
                            's1': s1_ids,
                            'target': y_id,
                        }
                        yield d
                    except:
                        continue

        count = items2tfrecord(items_gen(), tfrecord_file)
        return count

    @classmethod
    def load_tfrecord(cls, tfrecord_file, batch_size=128, index=None, shard=None):
        from qiznlp.common.tfrecord_utils import tfrecord2dataset, exist_tfrecord_file
        if not exist_tfrecord_file(tfrecord_file):
            return None, None
        feat_dct = {
            # 's1': tf.FixedLenFeature([50], tf.int64),
            's1': tf.VarLenFeature(tf.int64),
            'target': tf.FixedLenFeature([], tf.int64),
        }
        dataset, count = tfrecord2dataset(tfrecord_file, feat_dct, batch_size=batch_size, auto_pad=True, index=index, shard=shard)
        return dataset, count

    def get_signature_export_model(self):
        inputs_dct = {
            's1': self.s1,
            'dropout_rate': self.dropout_rate,
        }
        outputs_dct = {
            'y_prob': self.y_prob,
        }
        return inputs_dct, outputs_dct

    @classmethod
    def get_signature_load_pbmodel(cls):
        inputs_lst = ['s1', 'dropout_rate']
        outputs_lst = ['y_prob']
        return inputs_lst, outputs_lst

    @classmethod
    def from_pbmodel(cls, pbmodel_dir, sess):
        meta_graph_def = tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], pbmodel_dir)  # 从pb模型载入graph和variable,绑定到sess
        signature = meta_graph_def.signature_def  # 取出signature_def
        default_key = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY  # or 'chitchat_predict' 签名
        inputs_lst, output_lst = cls.get_signature_load_pbmodel()  # 类方法
        pb_dict = {}
        for k in inputs_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].inputs[k].name)  # 从signature中获取输入输出的tensor_name,并从graph中取出
        for k in output_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].outputs[k].name)
        model = cls(build_graph=False)  # 里面不再构造图
        for k, v in pb_dict.items():
            setattr(model, k, v)  # 绑定必要的输入输出到实例
        return model

    @classmethod
    def from_ckpt_meta(cls, ckpt_name, sess, graph):
        with graph.as_default():
            saver = tf.train.import_meta_graph(ckpt_name + '.meta', clear_devices=True)
            sess.run(tf.global_variables_initializer())

        model = cls(build_graph=False)  # 里面不再构造图
        # 绑定必要的输入输出到实例
        model.s1 = graph.get_tensor_by_name('s1:0')
        model.dropout_rate = graph.get_tensor_by_name('dropout_rate:0')
        # self.target = self.graph.get_tensor_by_name('target:0')
        model.y_prob = graph.get_tensor_by_name('y_prob:0')

        saver.restore(sess, ckpt_name)
        print(f':: restore success! {ckpt_name}')
        return model, saver
